/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * interp.c: Main interpretation loop for the SNAP virtual machine.
 *
 * $Id$
 */

#ifdef __KERNEL__
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/ip.h>
#include <net/route.h>
#include <snap/bytecode.h>
#include <snap/config.h>
#include <snap/d_printf.h>
#include <snap/dyncheck.h>
#include <snap/io.h>
#include <snap/myassert.h>
#include <snap/packet.h>
#include <snap/timers.h>
#include <snap/exception.h>
#include <snap/wassert.h>
#include <snap/snap_kern_iface.h>
#include <snap/services.h>
#else
#include <assert.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "config.h"
#include "bytecode.h"
#include "d_printf.h"
#include "deliver.h"
#include "dyncheck.h"
#include "io.h"
#include "memalloc.h"
#include "packet.h"
#include "router.h"
#include "timers.h"
#include "udpnet.h"
#include "exception.h"
#include "wassert.h"
#include "services.h"
#endif /* __KERNEL__ */

/************************************************************************/
/* macros for use in interp_packet */

/* Dynamic type checks */

#define DYNCHECK_IN_HEAP(ho)						\
  DYNCHECK((((void *)(ho) >= p->heap_min) &&				\
	    ((void *)(ho) < p->heap_max) &&				\
	    (((void *)(ho) + 						\
	      ((ho)->len + sizeof(heap_obj))) <= p->heap_max)) ||	\
	   (((void *)(ho) >= (void *)p->stack_max) &&			\
	    ((void *)(ho) < p->h_alloc_heap_max) &&			\
	    (((void *)(ho) + 						\
	      ((ho)->len + sizeof(heap_obj))) <= p->h_alloc_heap_max)))

#define DYNCHECK_ADDR_IN_HEAP(ho)				\
  DYNCHECK((((void *)(ho) >= p->heap_min) &&			\
	    ((void *)(ho) < p->heap_max) &&			\
	    (((void *)(ho) + sizeof(uint32) + 		\
	      sizeof(heap_obj)) <= p->heap_max)) ||		\
	   (((void *)(ho) >= (void *)p->stack_max) &&		\
	    ((void *)(ho) < p->h_alloc_heap_max) &&		\
	    (((void *)(ho) + sizeof(uint32) + 		\
	      sizeof(heap_obj)) <= p->h_alloc_heap_max)))

int heap_alloc(packet_t *p, int lenb, int flag, 
	       heap_obj **ho, int *hoffset) {
  int sizeb = lenb + sizeof(heap_obj);
  DYNCHECK(((void *)p->stack_max - sizeb) > (void *)p->sp);
  DYNCHECK(lenb <= MAX_HEAPOBJ_SZ);
  p->stack_max = (value_t *)((void *)p->stack_max - sizeb);
  *ho = (heap_obj *)((p)->stack_max);
  *hoffset = (void *)(p)->stack_max - (p)->heap_min;
  (*ho)->len = lenb;
  (*ho)->flag = flag;
  d_printf(150,"%s:%d: allocated %d-buf (f=%d) at %#x\n",
	   __FILE__,__LINE__,lenb,flag,(unsigned int)ho);
  return 0;
}

#ifdef ADDR_IN_HEAP
#define GET_BOXED(res,heap,v,t)					\
{								\
  heap_obj *ho1 = (heap_obj *)((heap) + GET_OFFS(v));		\
  DYNCHECK_ADDR_IN_HEAP(ho1);					\
  wassert(ho1->len == sizeof(t));				\
  (res) = *((t *)ho1->s);					\
}

#define GET_ADDR(res,h,v) GET_BOXED(res,h,v,uint32)
#define GET_FLOAT(res,h,v) GET_BOXED(res,h,v,float32)

#define SET_ADDR(val,a,p)				\
{ int hoffset;						\
  heap_obj *ho;						\
  if (!heap_alloc((p),sizeof(uint32),0,&ho,&hoffset)) {	\
    *((uint32 *)ho->s) = (a);				\
    SET_OFFS(val,hoffset);				\
  }							\
  else return -1;					\
}

#define SET_FLOAT(val,a,p)				\
{ int hoffset;						\
  heap_obj *ho;						\
  if (!heap_alloc((p),sizeof(float32),0,&ho,&hoffset)) {\
    *((float32 *)ho->s) = (a);				\
    SET_OFFS(val,hoffset);				\
  }							\
  else return -1;					\
}
#endif

/* Instruction cases */

#define INTOP(t,op,opname) {					\
  value_t *top1 = p->sp - 1;					\
  value_t *top2 = p->sp - 2;					\
  int result;							\
								\
  d_printf(100,"%s:%d: interp: %s\n",				\
           __FILE__,__LINE__,opname);				\
  DYNCHECK(top2 >= p->stack_min);				\
  DYNCHECK_TAG(*top1,INTV);					\
  DYNCHECK_TAG(*top2,INTV);					\
  result = (t)GET_INT(*top2) op (t)GET_INT(*top1);		\
  wassert((result <= MAX_VINT) && (result >= MIN_VINT));	\
  SET_TAG(*top2,INTV);						\
  SET_INT(*top2,result);					\
  p->sp--;							\
  p->pc++;							\
  break;							\
}

#define INTFLOATOP(op,opname) {					\
  value_t *top1 = p->sp - 1;					\
  value_t *top2 = p->sp - 2;					\
								\
  d_printf(100,"%s:%d: interp: %s\n",				\
           __FILE__,__LINE__,opname);				\
  DYNCHECK(top2 >= p->stack_min);				\
  if (GET_TAG(*top1) == INTV) {					\
    int result;							\
    DYNCHECK_TAG(*top2,INTV);					\
    result = (int)GET_INT(*top2) op (int)GET_INT(*top1);	\
    wassert((result <= MAX_VINT) && (result >= MIN_VINT));	\
    SET_TAG(*top2,INTV);					\
    SET_INT(*top2,result);					\
  } else {							\
    float32 v1, v2, result;					\
    DYNCHECK_TAG(*top1,FLOATV);					\
    DYNCHECK_TAG(*top2,FLOATV);					\
    GET_FLOAT(v1,p->heap_min,*top1);				\
    GET_FLOAT(v2,p->heap_min,*top2);				\
    result = v2 op v1;						\
    SET_TAG(*top2,FLOATV);					\
    SET_FLOAT(*top2,result,p);					\
  }								\
  p->sp--;							\
  p->pc++;							\
  break;							\
}

#define INTFLOATOPI(op,opname) {				\
  value_t *top1 = p->sp - 1;					\
  value_t *top2 = p->sp - 2;					\
								\
  d_printf(100,"%s:%d: interp: %s\n",				\
           __FILE__,__LINE__,opname);				\
  DYNCHECK(top2 >= p->stack_min);				\
  if (GET_TAG(*top1) == INTV) {					\
    int result;							\
    DYNCHECK_TAG(*top2,INTV);					\
    result = (int)GET_INT(*top2) op (int)GET_INT(*top1);	\
    wassert((result <= MAX_VINT) && (result >= MIN_VINT));	\
    SET_TAG(*top2,INTV);					\
    SET_INT(*top2,result);					\
  } else {							\
    float32 v1, v2;						\
    int result;							\
    DYNCHECK_TAG(*top1,FLOATV);					\
    DYNCHECK_TAG(*top2,FLOATV);					\
    GET_FLOAT(v1,p->heap_min,*top1);				\
    GET_FLOAT(v2,p->heap_min,*top2);				\
    result = v2 op v1;						\
    SET_TAG(*top2,INTV);					\
    SET_INT(*top2,result);					\
  }								\
  p->sp--;							\
  p->pc++;							\
  break;							\
}

#define INTIMMOP(t,op,opname) {					\
  value_t *top = p->sp - 1;					\
  int lit;							\
  int result;							\
								\
  d_printf(100,"%s:%d: interp: %s\n",				\
           __FILE__,__LINE__,opname);				\
  DYNCHECK(top >= p->stack_min);				\
  DYNCHECK_TAG(*top,INTV);					\
  GET_LIT(lit,INTV,*p->pc);					\
  result = (t)GET_INT(*top) op (t)lit;				\
  wassert((result <= MAX_VINT) && (result >= MIN_VINT));	\
  SET_TAG(*top,INTV);						\
  SET_INT(*top,result);						\
  p->pc++;							\
  break;							\
}

#define INTFLOATIMMOP(op,opname) {				\
  value_t *top = p->sp - 1;					\
								\
  d_printf(100,"%s:%d: interp: %s\n",				\
           __FILE__,__LINE__,opname);				\
  DYNCHECK(top >= p->stack_min);				\
  if (GET_TAG(*top) == INTV) {					\
    int lit;							\
    int result;							\
    GET_LIT(lit,INTV,*p->pc);					\
    result = GET_INT(*top) op lit;				\
    wassert((result <= MAX_VINT) && (result >= MIN_VINT));	\
    SET_TAG(*top,INTV);						\
    SET_INT(*top,result);					\
  } else {							\
    float32 f1, lit, result;					\
    value_t v = ZERO_VALUE_T;					\
    DYNCHECK_TAG(*top,FLOATV);					\
    COPY_LIT(v,FLOATV,*p->pc);					\
    GET_FLOAT(lit,p->heap_min,v);				\
    GET_FLOAT(f1,p->heap_min,*top);				\
    result = f1 op lit;						\
    SET_TAG(*top,FLOATV);					\
    SET_FLOAT(*top,result,p);					\
  }								\
  p->pc++;							\
  break;							\
}

#define INTFLOATIMMOPI(op,opname) {				\
  value_t *top = p->sp - 1;					\
								\
  d_printf(100,"%s:%d: interp: %s\n",				\
           __FILE__,__LINE__,opname);				\
  DYNCHECK(top >= p->stack_min);				\
  if (GET_TAG(*top) == INTV) {					\
    int lit;							\
    int result;							\
    GET_LIT(lit,INTV,*p->pc);					\
    result = GET_INT(*top) op lit;				\
    wassert((result <= MAX_VINT) && (result >= MIN_VINT));	\
    SET_TAG(*top,INTV);						\
    SET_INT(*top,result);					\
  } else {							\
    float32 f1, lit;						\
    int result;							\
    value_t v = ZERO_VALUE_T;					\
    DYNCHECK_TAG(*top,FLOATV);					\
    COPY_LIT(v,FLOATV,*p->pc);					\
    GET_FLOAT(lit,p->heap_min,v);				\
    GET_FLOAT(f1,p->heap_min,*top);				\
    result = f1 op lit;						\
    SET_TAG(*top,INTV);						\
    SET_INT(*top,result);					\
  }								\
  p->pc++;							\
  break;							\
}

#define FLOATIMMOP(op,opname) {			\
  value_t *top = p->sp - 1;			\
  value_t v = ZERO_VALUE_T;			\
  float32 result, f1, lit;			\
						\
  d_printf(100,"%s:%d: interp: %s\n",		\
           __FILE__,__LINE__,opname);		\
  DYNCHECK(top >= p->stack_min);		\
  DYNCHECK_TAG(*top,FLOATV);			\
  COPY_LIT(v,FLOATV,*p->pc);			\
  GET_FLOAT(lit,p->heap_min,v);			\
  GET_FLOAT(f1,p->heap_min,*top);		\
  result = f1 op lit;				\
  SET_TAG(*top,FLOATV);				\
  SET_FLOAT(*top,result,p);			\
  p->pc++;					\
  break;					\
}

#define FLOATIMMOPI(op,opname) {		\
  value_t *top = p->sp - 1;			\
  value_t v = ZERO_VALUE_T;			\
  float32 f1, lit;				\
  int result;					\
  d_printf(100,"%s:%d: interp: %s\n",		\
           __FILE__,__LINE__,opname);		\
  DYNCHECK(top >= p->stack_min);		\
  DYNCHECK_TAG(*top,FLOATV);			\
  COPY_LIT(v,FLOATV,*p->pc);			\
  GET_FLOAT(lit,p->heap_min,v);			\
  GET_FLOAT(f1,p->heap_min,*top);		\
  result = f1 op lit;				\
  SET_TAG(*top,INTV);				\
  SET_INT(*top,result);				\
  p->pc++;					\
  break;					\
}

#define EQOP(res,opname) {						\
  value_t *top1 = p->sp - 1;						\
  value_t *top2 = p->sp - 2;						\
  int result;								\
									\
  d_printf(100,"%s:%d: interp: %s\n",					\
	   __FILE__,__LINE__,opname);					\
  DYNCHECK(top2 >= p->stack_min);					\
  result = (res == eq_value(top1,top2,p));				\
  if (result == -1)							\
    return -1;								\
  SET_TAG(*top2,INTV);							\
  SET_INT(*top2,result);						\
  p->sp--;								\
  p->pc++;								\
  break;								\
}

#ifndef SMALL_INSTRS
#define EQOPI(res,opname) {						\
  value_t *top = p->sp - 1;						\
  int result;								\
  									\
  d_printf(100,"%s:%d: interp: %s\n",					\
	   __FILE__,__LINE__,opname);					\
  DYNCHECK(top >= p->stack_min);					\
  result = (res == eq_value(top,&p->pc->arg,p));			\
  SET_TAG(*top,INTV);							\
  SET_INT(*top,result);							\
  p->pc++;								\
  break;								\
}
#endif

/************************************************************************/

static int eq_value(value_t *v1, value_t *v2, packet_t *p) {
  int result;
  TAG_T tag1 = GET_TAG(*v1);
  switch(tag1) {
#ifdef SMALL_VALUES
    case INTV: 
    case TUPLEV: 
    case EXCV:
    case STRV: 
    default:
      result = (*v1 == *v2); break;
#else
    case EXCV:
    case INTV: 
    default:
      result = (GET_INT(*v1) == GET_INT(*v2)) && 
	       (tag1 == GET_TAG(*v2)); 
      break;
    case TUPLEV: 
    case STRV: 
      result = (GET_OFFS(*v1) == GET_OFFS(*v2)) &&
	       (tag1 == GET_TAG(*v2)); 
      break;
#endif
    case ADDRV: {
      uint32 addr1, addr2;
      if (tag1 == GET_TAG(*v2)) {
	GET_ADDR(addr1,p->heap_min,*v1);
	GET_ADDR(addr2,p->heap_min,*v2);
	result = (addr1 == addr2);
      }
      else
	result = 0;
      break;
    }
    case FLOATV: {
      float32 f1, f2;
      if (tag1 == GET_TAG(*v2)) {
	GET_FLOAT(f1,p->heap_min,*v1);
	GET_FLOAT(f2,p->heap_min,*v2);
	result = (f1 == f2); 
      }
      else
	result = 0;
      break;
    }
  }
  return result;
}

/* assumes packet is well-formed on entry; returns -1 on failure */
int interp_packet(packet_t *p) {
  int done;

  print_anti_timer(1,"interp_packet");
  done = 0;

  assert(p->pc >= p->code_min);

  d_printf(25,"%s:%d: interp_packet: %d instrs, %d stack vals, ep=%d, rb=%d\n",
	  __FILE__,__LINE__,p->code_max - p->code_min, p->sp -
	  p->stack_min, ntohl(p->hdr->entry_point), ntohs(p->hdr->rb));

  /* decrement resource bound on arrival */
  if (p->hdr->rb == 0) { /* don't have to convert, 0 == 0 whatever the
			    the byte order */
    /* out of resource bound after decrement */
    d_printf(25,"%s:%d: out of RB, dropping packet\n",__FILE__,__LINE__);
    return -1;
  }
  p->hdr->rb = htons(ntohs(p->hdr->rb) - 1);

  while(!done) {
    /* if we exactly fall off the end of the code, we can assume an
       implicit EXIT instruction */
    if (p->pc == p->code_max) {
      break;
    }

    DYNCHECK(p->pc < p->code_max);
    
    switch(GET_OP(*p->pc)) {
    case EXIT:
      d_printf(100,"%s:%d: interp: EXIT\n",__FILE__,__LINE__);
      done = 1; 
      break;

#ifndef SMALL_INSTRS
    case PUSH:
      d_printf(100,"%s:%d: interp: PUSH\n",__FILE__,__LINE__);
      DYNCHECK(p->sp < p->stack_max);
      COPY_VAL(*p->sp,p->pc->arg);
      p->sp++;
      p->pc++;
      break;
#else
#define PUSHOP(t,opname)					\
      d_printf(100,"%s:%d: interp: %s\n",__FILE__,__LINE__,	\
	       opname);						\
      DYNCHECK(p->sp < p->stack_max);				\
      COPY_LIT(*p->sp,t,*p->pc);				\
      p->sp++;							\
      p->pc++;							\
      break;

    case PINT: PUSHOP(INTV,"PINT");
    case PADDR: PUSHOP(ADDRV,"PADDR");
    case PTUP: PUSHOP(TUPLEV,"PTUP");
    case PEXC: PUSHOP(EXCV,"PEXC");
    case PSTR: PUSHOP(STRV,"PSTR");
    case PFLT: PUSHOP(FLOATV,"PFLT");
#endif

    case POP:
      d_printf(100,"%s:%d: interp: POP\n",__FILE__,__LINE__);
      DYNCHECK(p->sp > p->stack_min);
      p->sp--;
      p->pc++;
      break;

    case POPI: {
      int n;
      value_t *newsp;
      
      d_printf(100,"%s:%d: interp: POPI\n",__FILE__,__LINE__);
      GET_LIT(n,INTV,*p->pc);
      DYNCHECK(n >= 0);
      newsp = p->sp - n;
      DYNCHECK(newsp >= p->stack_min);
      p->sp = newsp;
      p->pc++;
      break;
    }

    case PULL: {
      int n;
      value_t *src;
      
      d_printf(100,"%s:%d: interp: PULL\n",__FILE__,__LINE__);
      GET_LIT(n,INTV,*p->pc);
      DYNCHECK(n >= 0);
      src = p->sp - (n + 1);
      DYNCHECK(src >= p->stack_min);
      DYNCHECK(p->sp < p->stack_max);
      COPY_VAL(*p->sp,*src);
      p->sp++;
      p->pc++;
      break;
    }

    case STORE: {
      int n;
      value_t *src;
      value_t *dst;
      
      d_printf(100,"%s:%d: interp: STORE\n",__FILE__,__LINE__);
      GET_LIT(n,INTV,*p->pc);
      DYNCHECK(n >= 0);
      src = p->sp - 1;
      dst = p->sp - (n + 1);
      DYNCHECK(src >= p->stack_min); /* also effectively checks that
					the stack is non-empty */
      COPY_VAL(*dst,*src);
      p->sp--;
      p->pc++;
      break;
    }

    case JI: {
      int jump;
     
      d_printf(100,"%s:%d: interp: JI\n",__FILE__,__LINE__);
      GET_LIT(jump,INTV,*p->pc);
      DYNCHECK(jump > 0);
      p->pc = p->pc + jump;
      break;
    }

    case PAJ: {
      value_t *top = p->sp - 1;
      int jump;
     
      d_printf(100,"%s:%d: interp: PAJ\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      GET_LIT(jump,INTV,*p->pc);
      DYNCHECK_TAG(*top,INTV);
      jump += GET_INT(*top);
      DYNCHECK(jump > 0);
      p->sp--;			/* pop */
      p->pc = p->pc + jump;
      break;
    }

    case TPAJ: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      int jump = 1;

      d_printf(100,"%s:%d: interp: TPAJ\n",__FILE__,__LINE__);
      DYNCHECK(top2 >= p->stack_min);
      DYNCHECK_TAG(*top1,INTV);
      if (GET_INT(*top1) == 0) {
	GET_LIT(jump,INTV,*p->pc);
	DYNCHECK_TAG(*top2,INTV);
	jump += GET_INT(*top2);
	DYNCHECK(jump > 0);
      }
      p->sp = p->sp - 2;		/* pop */
      p->pc = p->pc + jump;
      break;
    }

    case BEZ: {
      value_t *top = p->sp - 1;
      int jump = 1;

      d_printf(100,"%s:%d: interp: BEZ\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,INTV);
      if (GET_INT(*top) == 0) {
	GET_LIT(jump,INTV,*p->pc);
	DYNCHECK(jump > 0);
      }
      p->sp--;			/* pop */
      p->pc = p->pc + jump;
      break;
    }

    case BNE: {
      value_t *top = p->sp - 1;
      int jump = 1;
      
      d_printf(100,"%s:%d: interp: BNE\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,INTV);
      if (GET_INT(*top) != 0) {
	GET_LIT(jump,INTV,*p->pc);
	DYNCHECK(jump > 0);
      }
      p->sp--;			/* pop */
      p->pc = p->pc + jump;
      break;
    }

    case MKTUP: {
      int n, i, hoffset;
      value_t *newtop, *heapv;
      heap_obj *ho;
      
      d_printf(100,"%s:%d: interp: MKTUP\n",__FILE__,__LINE__);
      GET_LIT(n,INTV,*p->pc);
      DYNCHECK(n > 0);
      /* check for (n-2) trailing non-mktups */
      DYNCHECK((p->pc + ((n-2) > 0 ? (n-2) : 0) < p->code_max));
#ifndef NDYNCHECK
      for(i=1; i<n-1; i++) {
	/* check for non-mktup */
	DYNCHECK(GET_OP(p->pc[i]) != MKTUP);
      }
#endif
      /* check that the initial arguments are on the stack */
      newtop = (p->sp - n); /* where to store the tuple result */
      DYNCHECK(newtop >= p->stack_min);
      /* perform the allocation */
      if (!heap_alloc(p,n*sizeof(value_t),1,&ho,&hoffset)) {
	/* initialize the tuple entries */
	for (heapv = (value_t *)ho->s, p->sp -= n, i=0; 
	     i<n; 
	     i++, heapv++, p->sp++) {
	  *heapv = *p->sp;
	}
	/* store pointer on the stack */
	SET_TAG(*newtop,TUPLEV);
	SET_OFFS(*newtop,hoffset);
	p->sp = newtop + 1;
      }
      else /* allocation failed, quit */
	return -1;

      p->pc++;
      break;
    }

    case NTH:  {
      value_t *top1 = p->sp - 1;
      int n;
      heap_obj *ho;
      
      d_printf(100,"%s:%d: interp: NTH\n",__FILE__,__LINE__);
      DYNCHECK(top1 >= p->stack_min);
      DYNCHECK_TAG(*top1,TUPLEV);
      GET_LIT(n,INTV,*p->pc);
      DYNCHECK(n > 0);
      ho = (heap_obj *)(p->heap_min + GET_OFFS(*top1));
      DYNCHECK_IN_HEAP(ho);
      DYNCHECK((ho->len / 4) >= n);
      wassert((ho->len / 4) * 4 == ho->len);
      *top1 = ((value_t *)ho->s)[n-1];
      p->pc++;
      break;
    }

    case ISTUP: { /* unlike other ops, does not blow away top stack
		     value */
      value_t *top = p->sp - 1;
      int result;

      d_printf(100,"%s:%d: interp: ISTUP\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      result = (GET_TAG(*top) == TUPLEV);
      DYNCHECK(p->sp < p->stack_max);
      SET_TAG(*p->sp,INTV);
      SET_INT(*p->sp,result);
      p->sp++;
      p->pc++;
      break;
    }

    case LEN: {
      value_t *top = p->sp - 1;
      heap_obj *tup;

      d_printf(100,"%s:%d: interp: LEN\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,TUPLEV);
      tup = (heap_obj *)(p->heap_min + GET_OFFS(*top));
      DYNCHECK_IN_HEAP(tup);
      SET_TAG(*top,INTV);
      SET_INT(*top,tup->len / sizeof(value_t));
      p->pc++;
      break;      
    }

    /* equality/inequality are polymorphic */
    case EQ: EQOP(1,"EQ");
    case NEQ: EQOP(0,"NEQ");
#ifndef SMALL_INSTRS
    case EQI: EQOPI(1,"EQI");
    case NEQI: EQOPI(0,"NEQI");
    /* these could be on floats or on ints */
    case GTI: INTFLOATIMMOPI(>,"GTI");
    case GEQI: INTFLOATIMMOPI(>=,"GEQI");
    case LEQI: INTFLOATIMMOPI(<=,"LEQI");
    case LTI: INTFLOATIMMOPI(<,"LTI");
    case ADDI: INTFLOATIMMOP(+,"ADDI");
    case SUBI: INTFLOATIMMOP(-,"SUBI");
    case MULTI: INTFLOATIMMOP(*,"MULTI");
    case DIVI: INTFLOATIMMOP(/,"DIVI");
#else
#define EQOPI(res,t,opname)				\
{							\
  value_t v = ZERO_VALUE_T;				\
  int result;						\
  value_t *top = p->sp - 1;				\
  d_printf(100,"%s:%d: interp: %s\n",__FILE__,__LINE__,	\
	   opname);					\
  DYNCHECK(p->sp < p->stack_max);			\
  COPY_LIT(v,t,*p->pc);					\
  result = (res == eq_value(top,&v,p));			\
  if (result == -1)					\
    return -1;						\
  SET_TAG(*top,INTV);					\
  SET_INT(*top,result);					\
  p->pc++;						\
  break;						\
}
    case EQINT: EQOPI(1,INTV,"EQINT");
    case EQADR: EQOPI(1,ADDRV,"EQADR");
    case EQTUP: EQOPI(1,TUPLEV,"EQTUP");
    case EQEXC: EQOPI(1,EXCV,"EQEXC");
    case EQSTR: EQOPI(1,STRV,"EQSTR");
    case EQFLT: EQOPI(1,FLOATV,"EQFLT");

    case NQINT: EQOPI(0,INTV,"NQINT");
    case NQADR: EQOPI(0,ADDRV,"NQADR");
    case NQTUP: EQOPI(0,TUPLEV,"NQTUP");
    case NQEXC: EQOPI(0,EXCV,"NQEXC");
    case NQSTR: EQOPI(0,STRV,"NQSTR");
    case NQFLT: EQOPI(0,FLOATV,"EQFLT");

    /* floating point ops */
    case FGTI: FLOATIMMOPI(>,"GTI");
    case FGEQI: FLOATIMMOPI(>=,"GEQI");
    case FLEQI: FLOATIMMOPI(<=,"LEQI");
    case FLTI: FLOATIMMOPI(<,"LTI");
    case FADDI: FLOATIMMOP(+,"ADDI");
    case FSUBI: FLOATIMMOP(-,"SUBI");
    case FMULI: FLOATIMMOP(*,"MULTI");
    case FDIVI: FLOATIMMOP(/,"DIVI");

    /* integer ops */
    case GTI: INTIMMOP(int,>,"GTI");
    case GEQI: INTIMMOP(int,>=,"GEQI");
    case LEQI: INTIMMOP(int,<=,"LEQI");
    case LTI: INTIMMOP(int,<,"LTI");
    case ADDI: INTIMMOP(int,+,"ADDI");
    case SUBI: INTIMMOP(int,-,"SUBI");
    case MULTI: INTIMMOP(int,*,"MULTI");
    case DIVI: INTIMMOP(int,/,"DIVI");
#endif

    /* integer and/or floating point relational and arithmetic operators */
    case GT: INTFLOATOPI(>,"GT");
    case GEQ: INTFLOATOPI(>=,"GEQ");
    case LEQ: INTFLOATOPI(<=,"LEQ");
    case LT: INTFLOATOPI(<,"LT");
    case ADD: INTFLOATOP(+,"ADD");
    case SUB: INTFLOATOP(-,"SUB");
    case MULT: INTFLOATOP(*,"MULT");
    case DIV: INTFLOATOP(/,"DIV");

    case NEG: {
      value_t *top = p->sp - 1;

      d_printf(100,"%s:%d: interp: NEG\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      if (GET_TAG(*top) == INTV) {
	int result;
	result = -(GET_INT(*top));
	SET_TAG(*top,INTV);
	SET_INT(*top,result);
      } else {
	float32 v, result;
	DYNCHECK_TAG(*top,FLOATV);
	GET_FLOAT(v,p->heap_min,*top);
	result = -v;
	SET_TAG(*top,FLOATV);
	SET_FLOAT(*top,result,p);
      }
      p->pc++;
      break;
    }

    /* integer arithmetic operators */
    case MOD: INTOP(int,%,"MOD");
    case MODI: INTIMMOP(int,%,"MODI");
    case XOR: INTOP(int,^,"XOR");
    case XORI: INTIMMOP(int,^,"XORI");

    case NOT: {
      value_t *top = p->sp - 1;
      int result;

      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,INTV);
      result = !(GET_INT(*top));
      SET_TAG(*top,INTV);
      SET_INT(*top,result);
      p->pc++;
      break;
    }

    case LNOT: {
      value_t *top = p->sp - 1;
      int result;

      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,INTV);
      result = ~(GET_INT(*top));
      SET_TAG(*top,INTV);
      SET_INT(*top,result);
      p->pc++;
      break;
    }

    case AND: INTOP(int,&,"AND");
    case ANDI: INTIMMOP(int,&,"ANDI");
    case OR: INTOP(int,|,"OR");
    case ORI: INTIMMOP(int,|,"ORI");
    case LSHL: INTOP(int,<<,"LSHL");
    case LSHLI: INTIMMOP(int,<<,"LSHLI");
    case RSHL: INTOP(uint32,>>,"RSHL");
    case RSHLI: INTIMMOP(uint32,>>,"RSHLI");
    case RSHA: INTOP(int,>>,"RSHA");
    case RSHAI: INTIMMOP(int,>>,"RSHAI");

    case SNET: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      uint32 result, a1, a2;

      d_printf(100,"%s:%d: interp: SNET\n",__FILE__,__LINE__);
      DYNCHECK(top2 >= p->stack_min);
      DYNCHECK_TAG(*top1,ADDRV);
      DYNCHECK_TAG(*top2,ADDRV);
      GET_ADDR(a1,p->heap_min,*top1);
      GET_ADDR(a2,p->heap_min,*top2);
      result = a1 & a2;
      SET_TAG(*top2,ADDRV);
      SET_ADDR(*top2,result,p);
      p->sp--;
      p->pc++;
      break;
    }

    case SNETI: {
      value_t *top = p->sp - 1;
      uint32 a1, a2, result;

      d_printf(100,"%s:%d: interp: SNETI\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,ADDRV);
#ifndef ADDR_IN_HEAP
      DYNCHECK_TAG(p->pc->arg,ADDRV);
      GET_ADDR(a2,p->heap_min,p->pc->arg);
#else
      { 
	value_t v = ZERO_VALUE_T;
	COPY_LIT(v,ADDRV,*p->pc);
	GET_ADDR(a2,p->heap_min,v);
      }
#endif
      GET_ADDR(a1,p->heap_min,*top);
      result = a1 & a2;
      SET_TAG(*top,ADDRV);
      SET_ADDR(*top,result,p);
      p->pc++;
      break;
    }

    case BCAST: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      uint32 a1, a2, result;

      d_printf(100,"%s:%d: interp: BCAST\n",__FILE__,__LINE__);
      DYNCHECK(top2 >= p->stack_min);
      DYNCHECK_TAG(*top1,ADDRV);
      DYNCHECK_TAG(*top2,ADDRV);
      GET_ADDR(a1,p->heap_min,*top1);
      GET_ADDR(a2,p->heap_min,*top2);
      result = a1 | ~a2;
      SET_TAG(*top2,ADDRV);
      SET_ADDR(*top2,result,p);
      p->sp--;
      p->pc++;
      break;
    }

    case BCASTI: {
      value_t *top = p->sp - 1;
      uint32 a1, a2, result;

      d_printf(100,"%s:%d: interp: BCASTI\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,ADDRV);
#ifndef ADDR_IN_HEAP
      DYNCHECK_TAG(p->pc->arg,ADDRV);
      GET_ADDR(a2,p->heap_min,p->pc->arg);
#else
      { 
	value_t v = ZERO_VALUE_T;
	COPY_LIT(v,ADDRV,*p->pc);
	GET_ADDR(a2,p->heap_min,v);
      }
#endif
      GET_ADDR(a1,p->heap_min,*top);
      result = a1 | ~a2;
      SET_TAG(*top,ADDRV);
      SET_ADDR(*top,result,p);
      p->pc++;
      break;
    }

    case ISX: { /* unlike other ops, does not blow away top stack
		   value */
      value_t *top = p->sp - 1;
      int result;

      d_printf(100,"%s:%d: interp: ISX\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      result = (GET_TAG(*top) == EXCV);
      DYNCHECK(p->sp < p->stack_max);
      SET_TAG(*p->sp,INTV);
      SET_INT(*p->sp,result);
      p->sp++;
      p->pc++;
      break;
    }

    case GETRB: {
      d_printf(100,"%s:%d: interp: GETRB\n",__FILE__,__LINE__);
      DYNCHECK(p->sp < p->stack_max);
      SET_TAG(*p->sp,INTV);
      SET_INT(*p->sp, ntohs(p->hdr->rb));
      p->sp++;
      p->pc++;
      break;
    }

    case GETSRC: {
      d_printf(100,"%s:%d: interp: GETSRC\n",__FILE__,__LINE__);
      DYNCHECK(p->sp < p->stack_max);
      SET_TAG(*p->sp,ADDRV);
      SET_ADDR(*p->sp,p->hdr->src,p);
      p->sp++;
      p->pc++;
      break;
    }

    case GETDST: {
      d_printf(100,"%s:%d: interp: GETDST\n",__FILE__,__LINE__);
      DYNCHECK(p->sp < p->stack_max);
      SET_TAG(*p->sp,ADDRV);
      SET_ADDR(*p->sp,p->hdr->dst,p);
      p->sp++;
      p->pc++;
      break;
    }

    case GETSPT: {
      d_printf(100,"%s:%d: interp: GETSPT\n",__FILE__,__LINE__);
      DYNCHECK(p->sp < p->stack_max);
      SET_TAG(*p->sp,INTV);
      SET_INT(*p->sp,ntohs(p->hdr->sport));
      p->sp++;
      p->pc++;
      break;
    }

    case HERE: {
      d_printf(100,"%s:%d: interp: HERE\n",__FILE__,__LINE__);
#ifndef __KERNEL__
      assert(thishost != NULL);
#endif /* __KERNEL__ */
      DYNCHECK(p->sp < p->stack_max);
      SET_TAG(*p->sp,ADDRV);
#ifdef __KERNEL__
      SET_ADDR(*p->sp,((struct rtable *)(p->skb->dst))->rt_dst,p);
#else
      SET_ADDR(*p->sp,thisaddr.s_addr,p);
#endif /* __KERNEL__ */
      p->sp++;
      p->pc++;
      break;
    }

    case ISHERE: {
      value_t *top = p->sp - 1;
      uint32 addr;
      int result;

#ifndef __KERNEL__
      assert(thishost != NULL);
#endif /* __KERNEL__ */
      d_printf(100,"%s:%d: interp: ISHERE\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,ADDRV);
      GET_ADDR(addr,p->heap_min,*top);
      print_timer(1,"interp_packet");
#ifdef __KERNEL__
      result = ishere(addr,p->skb);
#else
      result = udp_ishere(addr);
#endif /* __KERNEL__ */
      print_anti_timer(1,"interp_packet");
      SET_TAG(*top,INTV);
      SET_INT(*top,result);
      p->pc++;
      break;
    }

    case ROUTE: {
      value_t *top = p->sp - 1;
      uint32 *ret,addr;
#ifdef __KERNEL__
      struct rtable *rt;
      u32 hop = 0;
#endif /* __KERNEL__ */

      d_printf(100,"%s:%d: interp: ROUTE\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,ADDRV);
      GET_ADDR(addr,p->heap_min,*top);
      print_timer(1,"interp_packet");
#ifdef __KERNEL__
      if (next_hop(&rt,addr,p->skb->nh.iph->tos)) {
	/* no route found */
	hop = 0;
	ret = &hop;
      } else {
	hop = rt->rt_dst;
	ip_rt_put(rt);
	ret = &hop;
      }
#else
      ret = nexthop(addr);
#endif /* __KERNEL__ */
      print_anti_timer(1,"interp_packet");
      DYNCHECK(ret != NULL);
      if (*ret != 0) {
	SET_TAG(*top,ADDRV);
	SET_ADDR(*top,*ret,p);
      }
      else {
	SET_TAG(*top,EXCV);
	SET_INT(*top,E_NO_ROUTE);
      }
      p->pc++;
      break;
    }

    case SEND: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      value_t *top3 = p->sp - 3;
      value_t *top4 = p->sp - 4;
      uint32 dest;
      int rb;

      d_printf(100,"%s:%d: interp: SEND\n",__FILE__,__LINE__);
      DYNCHECK(top4 >= p->stack_min);
      DYNCHECK_TAG(*top1,ADDRV);
      GET_ADDR(dest,p->heap_min,*top1);
      DYNCHECK_TAG(*top2,INTV);
      rb = GET_INT(*top2);
      p->sp -= 3;		/* pop; leave room for return value */

      if (rb > 0) {
	if (rb <= ntohs(p->hdr->rb)) {
	  uint32 srccache,dstcache;
	  short rb_cache;
	  int ep,st_amt;
      
	  DYNCHECK_TAG(*top3,INTV);
	  st_amt = GET_INT(*top3);
	  DYNCHECK_TAG(*top4,INTV);
	  ep = GET_INT(*top4);
	  dstcache = p->hdr->dst;
	  srccache = p->hdr->src;
	  rb_cache = ntohs(p->hdr->rb);
	  p->hdr->dst = dest;
#ifdef __KERNEL__
	  p->hdr->src = ((struct rtable *)p->skb->dst)->rt_dst;
#else
	  p->hdr->src = thisaddr.s_addr;
#endif /* __KERNEL__ */
	  p->hdr->rb = htons(rb);
	  p->hdr->entry_point = htons(ep);
	  p->sp--;
	  print_timer(1,"interp_packet");
#ifdef __KERNEL__
	  snap_send_packet(p,st_amt,dest);
#else
	  udp_send_packet(p,st_amt,dest);
#endif /* __KERNEL__ */
	  print_anti_timer(1,"interp_packet");
	  p->sp++;
	  p->hdr->dst = dstcache;
	  p->hdr->src = srccache;
	  p->hdr->rb = htons(rb_cache - rb);
	  SET_TAG(*top4,INTV);
	  SET_INT(*top4,0);
	}
	/* not enough resource bound */
	else {
	  SET_TAG(*top4,EXCV);
	  SET_INT(*top4,E_NOT_ENOUGH_RB);
	}
      }
      /* non-positive resource bound */
      else {
	SET_TAG(*top4,EXCV);
	SET_INT(*top4,E_NON_POSITIVE_RB);
      }
      p->pc++;
      break;
    }

    case SENDCK: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      value_t *top3 = p->sp - 3;
      uint32 dest;
      int rb;

      d_printf(100,"%s:%d: interp: SENDCK\n",__FILE__,__LINE__);
      DYNCHECK(top3 >= p->stack_min);
      DYNCHECK_TAG(*top1,ADDRV);
      GET_ADDR(dest,p->heap_min,*top1);
      DYNCHECK_TAG(*top2,INTV);
      rb = GET_INT(*top2);

      p->sp -= 2;		/* pop; leave room for return value */

      if (rb > 0) {
	if (rb <= ntohs(p->hdr->rb)) {
	  uint32 dstcache,srccache;
	  short rb_cache;
	  int ep,i,st_amt;
	  heap_obj *chunk, *hbindings;
	  value_t *bindings;

	  DYNCHECK_TAG(*top3,TUPLEV);
	  chunk = (heap_obj *)(p->heap_min + GET_OFFS(*top3));
	  DYNCHECK_IN_HEAP(chunk);
	  DYNCHECK(chunk->len == (2*sizeof(value_t)));

	  DYNCHECK_TAG(((value_t *)chunk->s)[0],INTV);
	  ep = GET_INT(((value_t *)chunk->s)[0]);
	  p->sp--;

	  /* no arguments */
	  if (GET_TAG(((value_t *)chunk->s)[1]) != TUPLEV)
	    st_amt = 0;
	  /* copy arguments to the stack */
	  else {
	    hbindings = (heap_obj *)(p->heap_min + 
				     GET_OFFS(((value_t *)chunk->s)[1]));
	    DYNCHECK_IN_HEAP(hbindings);
	    st_amt = hbindings->len / sizeof(value_t);
	    DYNCHECK((p->sp + st_amt) < p->stack_max);
	    bindings = (value_t *)hbindings->s;
	    for (i=0; i<st_amt; i++, p->sp++, bindings++)
	      *p->sp = *bindings;
	  }
	  dstcache = p->hdr->dst;
	  srccache = p->hdr->src;
	  rb_cache = ntohs(p->hdr->rb);
	  p->hdr->dst = dest;
#ifdef __KERNEL__
	  p->hdr->src = ((struct rtable *)p->skb->dst)->rt_dst;
#else
	  p->hdr->src = thisaddr.s_addr;
#endif /* __KERNEL__ */
	  p->hdr->rb = htons(rb);
	  p->hdr->entry_point = htons(ep);
	  print_timer(1,"interp_packet");
#ifdef __KERNEL__
	  snap_send_packet(p,st_amt,dest);
#else
	  udp_send_packet(p,st_amt,dest);
#endif /* __KERNEL__ */
	  print_anti_timer(1,"interp_packet");
	  /* remove args from the stack */
	  p->sp -= (st_amt-1);
	  p->hdr->dst = dstcache;
	  p->hdr->src = srccache;
	  p->hdr->rb = htons(rb_cache - rb);
	  SET_TAG(*top3,INTV);
	  SET_INT(*top3,0);
	}
	/* not enough resource bound */
	else {
	  SET_TAG(*top3,EXCV);
	  SET_INT(*top3,E_NOT_ENOUGH_RB);
	}
      }
      /* non-positive resource bound */
      else {
	SET_TAG(*top3,EXCV);
	SET_INT(*top3,E_NON_POSITIVE_RB);
      }
      p->pc++;
      break;
    }

    case HOP: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      value_t *top3 = p->sp - 3;
      value_t *top4 = p->sp - 4;
      uint32 dest;
      int rb;

      d_printf(100,"%s:%d: interp: HOP\n",__FILE__,__LINE__);
      DYNCHECK(top4 >= p->stack_min);
      DYNCHECK_TAG(*top1,ADDRV);
      GET_ADDR(dest,p->heap_min,*top1);
      DYNCHECK_TAG(*top2,INTV);
      rb = GET_INT(*top2);
      p->sp -= 3;		/* pop; leave room for return value */

      if (rb > 0) {
	if (rb <= ntohs(p->hdr->rb)) {
	  short rb_cache;
	  int ep,st_amt;
      
	  DYNCHECK_TAG(*top3,INTV);
	  st_amt = GET_INT(*top3);
	  DYNCHECK_TAG(*top4,INTV);
	  ep = GET_INT(*top4);

	  rb_cache = ntohs(p->hdr->rb);
	  p->hdr->entry_point = htons(ep);
	  p->hdr->rb = htons(rb);
	  p->sp--;
	  print_timer(1,"interp_packet");
#ifdef __KERNEL__
	  snap_send_packet(p,st_amt,dest);
#else
	  udp_send_packet(p,st_amt,dest);
#endif /* __KERNEL__ */
	  print_anti_timer(1,"interp_packet");
	  p->sp++;
	  p->hdr->rb = htons(rb_cache - rb);
	  SET_TAG(*top4,INTV);
	  SET_INT(*top4,0);
	}
	/* not enough resource bound */
	else {
	  SET_TAG(*top4,EXCV);
	  SET_INT(*top4,E_NOT_ENOUGH_RB);
	}
      }
      /* non-positive resource bound */
      else {
	SET_TAG(*top4,EXCV);
	SET_INT(*top4,E_NON_POSITIVE_RB);
      }
      p->pc++;
      break;
    }

    case HOPCK: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      value_t *top3 = p->sp - 3;
      uint32 dest;
      int rb;

      d_printf(100,"%s:%d: interp: HOPCK\n",__FILE__,__LINE__);
      DYNCHECK(top3 >= p->stack_min);
      DYNCHECK_TAG(*top1,ADDRV);
      GET_ADDR(dest,p->heap_min,*top1);
      DYNCHECK_TAG(*top2,INTV);
      rb = GET_INT(*top2);
      p->sp -= 2;		/* pop */

      if (rb > 0) {
	if (rb <= ntohs(p->hdr->rb)) {
	  heap_obj *chunk, *hbindings;
	  value_t *bindings;
	  int st_amt, ep, i;
	  short rb_cache;

	  DYNCHECK_TAG(*top3,TUPLEV);
	  chunk = (heap_obj *)(p->heap_min + GET_OFFS(*top3));
	  DYNCHECK_IN_HEAP(chunk);
	  DYNCHECK(chunk->len == (2*sizeof(value_t)));

	  DYNCHECK_TAG(((value_t *)chunk->s)[0],INTV);
	  ep = GET_INT(((value_t *)chunk->s)[0]);

	  /* no arguments */
	  if (GET_TAG(((value_t *)chunk->s)[1]) != TUPLEV)
	    st_amt = 0;
	  /* copy arguments to the stack */
	  else {
	    hbindings = (heap_obj *)(p->heap_min + 
				     GET_OFFS(((value_t *)chunk->s)[1]));
	    DYNCHECK_IN_HEAP(hbindings);
	    st_amt = hbindings->len / sizeof(value_t);
	    DYNCHECK((p->sp + st_amt) < p->stack_max);
	    bindings = (value_t *)hbindings->s;
	    for (i=0; i<st_amt; i++, p->sp++, bindings++)
	      *p->sp = *bindings;
	  }

	  rb_cache = ntohs(p->hdr->rb);
	  p->hdr->entry_point = htons(ep);
	  p->hdr->rb = htons(rb);
	  print_timer(1,"interp_packet");
#ifdef __KERNEL__
	  snap_send_packet(p,st_amt,dest);
#else
	  udp_send_packet(p,st_amt,dest);
#endif /* __KERNEL__ */
	  print_anti_timer(1,"interp_packet");
	  /* remove args from the stack */
	  p->sp -= (st_amt-1);
	  p->hdr->rb = htons(rb_cache - rb);
	  SET_TAG(*top3,INTV);
	  SET_INT(*top3,0);
	}
	/* not enough resource bound */
	else {
	  SET_TAG(*top3,EXCV);
	  SET_INT(*top3,E_NOT_ENOUGH_RB);
	}
      }
      /* non-positive resource bound */
      else {
	SET_TAG(*top3,EXCV);
	SET_INT(*top3,E_NON_POSITIVE_RB);
      }
      p->pc++;
      break;
    }


    case FORW: {
      int herep;

      print_timer(1,"interp_packet");
#ifdef __KERNEL__
      herep = ishere(p->hdr->dst, p->skb);
#else
      herep = udp_ishere(p->hdr->dst);
#endif /* __KERNEL__ */
      print_anti_timer(1,"interp_packet");
      if (herep) {
	d_printf(100,"%s:%d: interp: FORW (at dest)\n",__FILE__,__LINE__);
      } else {
#ifdef __KERNEL__
	struct rtable *rt = NULL;
	u32 hopaddr;
#endif /* __KERNEL__ */
	int *hop;
	d_printf(100,"%s:%d: interp: FORW (moving on)\n",__FILE__,__LINE__);
	print_timer(1,"interp_packet");
#ifdef __KERNEL__
	if (next_hop(&rt,p->hdr->dst,p->skb->nh.iph->tos)) {
	  hop = NULL;
	} else {
	  hopaddr = rt->rt_dst;
	  hop = &hopaddr;
	  ip_rt_put(rt);
	}
#else
	hop = nexthop(p->hdr->dst);
#endif /* __KERNEL__ */	
	print_anti_timer(1,"interp_packet");
	if (hop == NULL) {  /* no route */
#ifdef __KERNEL__
	  printk(KERN_WARNING "No route to host ");
	  printk_addr(p->hdr->dst);
	  printk("; dropping packet\n");
#else
	  fprintf(stderr,"No route to host ");
	  fprintf_addr(stderr,p->hdr->dst);
	  fprintf(stderr,"; dropping packet\n");
#endif /* __KERNEL__ */
	}
	else {
	  DYNCHECK(ntohs(p->hdr->rb) > 0); /* OK to do dyncheck here
					      because packet will just
					      get dropped anyway */
	  print_timer(1,"interp_packet");
#ifdef __KERNEL__
	  snap_send_packet(p, p->sp - p->stack_min, hopaddr);
#else
	  udp_send_packet(p, p->sp - p->stack_min,
			  (*hop == 0 ? p->hdr->dst : *hop));
#endif /* __KERNEL__ */
	  print_anti_timer(1,"interp_packet");
	  p->hdr->rb = htons(0);
	}
	done = 1;
      }
      p->pc++;
      break;
    }

    case FORWTO: {
      value_t *top = p->sp - 1;
#ifdef __KERNEL__
      struct rtable *rt = NULL;
      u32 hopaddr;
#endif /* __KERNEL__ */
      uint32 *hop;
      int herep;
      
      d_printf(100,"%s:%d: interp: FORWTO",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      DYNCHECK_TAG(*top,ADDRV);
      GET_ADDR(p->hdr->dst,p->heap_min,*top);
#ifdef __KERNEL__
      p->hdr->src = ((struct rtable *)p->skb->dst)->rt_dst;
#else
      p->hdr->src = thisaddr.s_addr;
#endif /* __KERNEL__ */
      p->sp--;			
      print_timer(1,"interp_packet");
#ifdef __KERNEL__
      herep = ishere(p->hdr->dst,p->skb);
#else
      herep = udp_ishere(p->hdr->dst);
#endif /* __KERNEL__ */
      print_anti_timer(1,"interp_packet");
      if (!herep) {
	d_printf(150,"%s:%d: interp: FORWTO (moving on)\n",__FILE__,__LINE__);
	print_timer(1,"interp_packet");
#ifdef __KERNEL__
	if (next_hop(&rt,p->hdr->dst,p->skb->nh.iph->tos)) {
	  hop = NULL;
	} else {
	  hopaddr = rt->rt_dst;
	  hop = &hopaddr;
	  ip_rt_put(rt);
	}
#else
	hop = nexthop(p->hdr->dst);
#endif /* __KERNEL__ */
	print_anti_timer(1,"interp_packet");
	if (hop == NULL) {
#ifdef __KERNEL__
	  printk(KERN_WARNING "No route to host ");
	  printk_addr(p->hdr->dst);
	  printk("; dropping packet\n");
#else
	  fprintf(stderr,"No route to host ");
	  fprintf_addr(stderr,p->hdr->dst);
	  fprintf(stderr,"; droppping packet\n");
#endif /* __KERNEL__ */
	}
	else {
	  DYNCHECK(ntohs(p->hdr->rb) > 0); /* OK to do dyncheck here
					      because packet will just
					      get dropped anyway */
	  print_timer(1,"interp_packet");
#ifdef __KERNEL__
	  snap_send_packet(p, p->sp - p->stack_min, hopaddr);
#else
	  if (*hop == 0)
	    udp_send_packet(p, p->sp - p->stack_min, p->hdr->dst);
	  else
	    udp_send_packet(p, p->sp - p->stack_min, *hop);
#endif /* __KERNEL__ */
	  print_anti_timer(1,"interp_packet");
	  p->hdr->rb = htons(0);
	}
	done = 1;
      } else {
	d_printf(150,"%s:%d: interp: FORWTO (at dest)\n",__FILE__,__LINE__);
      }
      p->pc++;
      break;
    }

    case DEMUX: {
      value_t *top1 = p->sp - 1;
      value_t *top2 = p->sp - 2;
      int pnum = 0;
      char *s;
      int slen = 0;
      heap_obj *ho;

      d_printf(100,"%s:%d: interp: DEMUX\n",__FILE__,__LINE__);
      DYNCHECK(top2 >= p->stack_min);
      DYNCHECK_TAG(*top1,INTV);
      pnum = GET_INT(*top1);
      DYNCHECK(pnum > 0);
      DYNCHECK_TAG(*top2,STRV);
      ho = (heap_obj *)(p->heap_min + GET_OFFS(*top2));
      DYNCHECK_IN_HEAP(ho);
      s = (char *)&(ho->s);
      slen = ho->len;
      d_printf(150,"%s:%d: deliver %d bytes to port %d\n",
	       __FILE__,__LINE__,slen,pnum);
      print_timer(1,"interp_packet");
#ifdef __KERNEL__
      snap_demux(pnum,p->skb,s,slen);
#else
      deliver(pnum,s,slen);
#endif /* __KERNEL__ */
      print_anti_timer(1,"interp_packet");
      p->sp -= 2;		/* pop */
      p->pc++;
      done = 1; 			/* exit after one demux */
      break;
    }

    case DEMUXI: {
      value_t *top = p->sp - 1;
      int pnum = 0;
      char *s;
      int slen = 0;
      heap_obj *ho;

      d_printf(100,"%s:%d: interp: DEMUXI\n",__FILE__,__LINE__);
      DYNCHECK(top >= p->stack_min);
      GET_LIT(pnum,INTV,*p->pc);
      DYNCHECK(pnum > 0);
      DYNCHECK_TAG(*top,STRV);
      ho = (heap_obj *)(p->heap_min + GET_OFFS(*top));
      DYNCHECK_IN_HEAP(ho);
      s = (char *)&(ho->s);
      slen = ho->len;
      d_printf(150,"%s:%d: deliver %d bytes to port %d\n",
	       __FILE__,__LINE__,slen,pnum);
      print_timer(1,"interp_packet");
#ifdef __KERNEL__
      snap_demux(pnum,p->skb,s,slen);
#else
      deliver(pnum,s,slen);
#endif /* __KERNEL__ */
      print_anti_timer(1,"interp_packet");
      p->sp--; 			/* pop */
      p->pc++;
      done = 1;			/* exit after one demux */
      break;
    }

#ifndef NDEBUG
    case PRINT: {
      value_t *top = p->sp - 1;
      
      DYNCHECK(top >= p->stack_min);
      d_printf(100,"%s:%d: interp: PRINT\n",__FILE__,__LINE__);
#ifdef __KERNEL__
      printk(KERN_WARNING "print: ");
      printk_value(p->heap_min,top);
      printk("\n");
#else
      fprintf_value(p->heap_min,stdout,top);
      printf("\n");
      fflush(stdout);
#endif /* __KERNEL__ */
      p->pc++;
      break;
    }
#endif  /* !NDEBUG */

#ifdef SERVICES
    case SVCV: {
      int offs, retc;
      heap_obj *ho;
      char *s;

      DYNCHECK(p->sp < p->stack_max);
      d_printf(100,"%s:%d: interp: SVCV\n",__FILE__,__LINE__);

      GET_LIT(offs,STRV,*p->pc);
      ho = (heap_obj *)(p->heap_min + offs);
      DYNCHECK_IN_HEAP(ho);
      s = (char *)&(ho->s);
      retc = lookup_value(p,s,p->sp);
      if (retc != 0) { /* not found */
	SET_TAG(*p->sp,EXCV);
	SET_INT(*p->sp,E_SERVICE_NOT_PRESENT);
      }
      p->sp++;
      p->pc++;
      break;
    }

    case CALLS: {
      int offs, retc;
      heap_obj *ho;
      char *s;

      DYNCHECK(p->sp < p->stack_max);
      d_printf(100,"%s:%d: interp: CALLS\n",__FILE__,__LINE__);

      GET_LIT(offs,STRV,*p->pc);
      ho = (heap_obj *)(p->heap_min + offs);
      DYNCHECK_IN_HEAP(ho);
      s = (char *)&(ho->s);
      retc = call_service(p,s,p->sp);
      if (retc != 0) { /* not found */
	SET_TAG(*p->sp,EXCV);
	SET_INT(*p->sp,E_SERVICE_NOT_PRESENT);
      }
      p->sp++;
      p->pc++;
      break;
    }
#endif /* SERVICES */

    default:
#ifdef __KERNEL__
      printk(KERN_WARNING "%s: unknown opcode: 0x%x\n",__FILE__,GET_OP(*p->pc));
      return -1;
#else
      fprintf(stderr,"%s: unknown opcode: 0x%x\n",__FILE__,GET_OP(*p->pc));
      fflush(stderr);
      exit(1);
#endif /* __KERNEL__ */
    }
  }
  print_timer(1,"interp_packet");
  return 0;
}
