/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * services.c : management and invocation of node-resident services
 *
 * $Id$
 */
#include <assert.h>
#include "bytecode.h"
#include "d_printf.h"
#include "hashtable.h"
#include "memalloc.h"
#include "io.h"
#include "interp.h"

#ifdef CONFIG_IP_SNAP_SVCS
#define SERVICES
#endif

#ifdef SERVICES
static hash_table_t *svc_val_tab = NULL;
static hash_table_t *svc_fun_tab = NULL;
static buffer_t *svc_heap = NULL;
static int svc_heap_alloc_offs = 0;

typedef value_t (*svc_fun)(packet_t *p);

#define DEF_SVC_TAB_SZ 53

#define mystrcmp strcmp

static int copy_value(packet_t *p, value_t *srcv, value_t *destv) {
  switch (GET_TAG(*srcv)) {
    case INTV:
    case EXCV:
    default:
      COPY_VAL(*destv,*srcv); 
      return 0;
    case ADDRV:
    case FLOATV:
#ifndef ADDR_IN_HEAP
      COPY_VAL(*destv,*srcv); 
      return 0;
#endif
    case STRV:
    case TUPLEV: {
      /* copy the object to the packet heap */
      heap_obj *pkt_ho;
      int pkt_offs, svc_offs = GET_OFFS(*srcv);
      heap_obj *svc_ho = (heap_obj *)(svc_heap->s + svc_offs);
      assert(svc_ho->flag == 1); /* XXX no support for tuples yet */
      if (!heap_alloc(p,svc_ho->len,svc_ho->flag,
		      &pkt_ho,&pkt_offs)) {
	memcpy(pkt_ho->s,svc_ho->s,svc_ho->len);
	SET_TAG(*destv,GET_TAG(*srcv));
	SET_OFFS(*destv,pkt_offs);
	return 0;
      }
      else
	return -1; /* no room to copy the result */
    }
  }
}

/* looks up a value in the service symbol table */
int lookup_value(packet_t *p, char *name, value_t *retv) {
  value_t *v;

  assert(p != NULL);
  assert(svc_val_tab != NULL);
  assert(name != NULL);

  v = (value_t *)ht_lookup(svc_val_tab,(void *)name);
  if (v != NULL)
    return copy_value(p,v,retv);
  else
    return -1; /* value not found */
}


/* looks up a service function, calls the function, and returns
   the result.  Assumes the service will do its own argument
   checking */
int call_service(packet_t *p, char *name, value_t *retv) {
  void *pf;
  value_t v;

  assert(p != NULL);
  assert(svc_fun_tab != NULL);
  assert(name != NULL);

  if ((pf = ht_lookup(svc_fun_tab,(void *)name)) != NULL) {
    v = ((svc_fun)pf)(p);
    return copy_value(p,&v,retv);
  }
  else
    return -1; /* value not found */
}
    

/*********************************************************************/

/* test service */
static value_t test_svc(packet_t *p) {
  value_t v = ZERO_VALUE_T;
  SET_TAG(v,INTV);
  SET_INT(v,11);
  return v;
}

static value_t test_val;
  
/* inserts a service function into the heap */
int register_svc_fun(char *name, svc_fun f) {
  assert(name != NULL);
  assert(f != NULL);

  ht_insert(svc_fun_tab,name,f);
  return 0;
}

/* initializes the service table with the "core" values; stores the
   heap objects in the provided buffer; offsets are made relative
   to the given heap */
void init_services(buffer_t *b) {
  
  assert(svc_val_tab == NULL);
  assert(svc_fun_tab == NULL);
  assert(svc_heap == NULL);
  assert(b != NULL);
  assert(b->s != NULL);

  /* initialize the service data structures */
  svc_val_tab = ht_create(DEF_SVC_TAB_SZ,mystrcmp,hash_string);
  svc_fun_tab = ht_create(DEF_SVC_TAB_SZ,mystrcmp,hash_string);
  svc_heap = b;
  svc_heap_alloc_offs = 0;

  /* test service */
  register_svc_fun("testSvc",test_svc);
  /* test value */
  SET_TAG(test_val,INTV);
  SET_INT(test_val,12);
  ht_insert(svc_val_tab,"testVal",&test_val);
}
#endif /* SERVICES */



#ifdef FOO
  heap_obj *ho;
  int offs;
  int i, sizeb, lenb;
  char *bufptr = b->s;
  int bufsz = b->lenb;
  value_t *v;

    /* allocate the heap object */
    sizeb = lenb = strlen(exns[i]) + 1;
    if ((sizeb & 0x3) != 0) {
      sizeb = (sizeb & ~0x3) + 4;
    }
    if (((bufsz - (sizeb + sizeof(heap_obj))) < 0) ||
	(sizeb > MAX_HEAPOBJ_SZ))  {
      fprintf(stderr,"%s:%d: heap alloc failed, buffer is too small\n",
	      __FILE__,__LINE__);
      fflush(stderr);
      exit(1);
    }
    ho = (heap_obj *)bufptr;
    offs = (void *)bufptr - (void *)b->s;
    assert(offs >= 0);
    assert(offs <= MAX_VINT);

    bufptr += (sizeb + sizeof(heap_obj));
    
    d_printf(150,"%s:%d: alloc svcval %s at @%#x, offs = %d\n",
	     __FILE__,__LINE__, exns[i], bufptr, offs);
    
    bzero(ho,sizeb + sizeof(heap_obj));
    ho->len = lenb;
    ho->flag = 0;
    memcpy((char *)(&(ho->s)),exns[i],lenb);

    /* create the value */
    memalloc(v,value_t *,sizeof(value_t));
    SET_TAG(*v,EXCV);
    SET_OFFS(*v,offs);

    /* insert it in the table */
    ht_insert(svc_val_tab,exns[i],v);
  }
}
#endif
