/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * snapi.c : SNAP interpreter. Non-network version of snapd, useful
 *   for testing per-node behavior of SNAP programs. Reads a binary
 *   packet on stdin, executes the program, and on exit dumps the
 *   residual packet on stdout.
 *
 * $Id$
 */

#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <netinet/in.h>
#include "packet.h"
#include "bytecode.h"
#include "memalloc.h"
#include "io.h"
#include "interp.h"
#include "router.h"
#include "udpnet.h"
#include "services.h"
#include "d_printf.h"

extern char *basename(const char *);

#define SVC_BUF_SZ 5012

void parse_cmdline(int argc,char **argv);
int infd;
int outfd = 1;

int main(int argc,char **argv) {
  buffer_t inbuf;
#ifdef SERVICES
  buffer_t service_heap;
#endif
  packet_t *p;
  buffer_t outbuf;
  int packet_lenb;

  d_printf(10,"%s:%d: starting up\n",__FILE__,__LINE__);

  parse_cmdline(argc,argv);

  read_routes();
  init_udpnet();
  packet_lenb = file_to_str(infd,&inbuf);

#ifdef SERVICES
  memalloc(service_heap.s,char *,SVC_BUF_SZ);
  service_heap.lenb = SVC_BUF_SZ;
  assert(service_heap.s > inbuf.s);
#endif

  p = unmarshal_packet(inbuf.s,packet_lenb,inbuf.lenb);
  if (p != NULL) {
#ifdef SERVICES
    init_services(&service_heap);
#endif
    /* add 1 to rb to allow tests to run */
    p->hdr->rb = htons(ntohs(p->hdr->rb) + 1);
    if (!interp_packet(p)) {
      marshal_packet(p, p->sp - p->stack_min, &outbuf);
      write(outfd,outbuf.s,outbuf.lenb);
      return(0);
    }
    else
      return(1);
  }
  else
    return(1);
}

void usage(int argc, char **argv) {
  printf("usage: %s [-l?] [-o outfile] infile\n",basename(argv[0]));
  printf("\t-? : print this help\n");
  printf("\t-o : set the output file name (default=stdout)\n");
}

void parse_cmdline(int argc, char **argv) {
  char *opts = "?o:";
  char c;
  int args_expected = 1;
  int args_processed = 0;
  int argidx;
  char *outfilename;
  char *infilename;

  while((c = getopt(argc,argv,opts)) != EOF) {
    switch(c) {
    case '?':
      usage(argc,argv);
      exit(1);
    case 'o':
      outfilename = optarg;
      if (strcmp(outfilename,"-") == 0) {
	outfd = 1;
      } else {
	outfd = open(outfilename,O_WRONLY|O_CREAT|O_TRUNC,
		     S_IREAD|S_IWRITE|S_IRGRP|S_IROTH);
	if (outfd == -1) {
	  fprintf(stderr,"%s: unable to open file \"%s\" for output\n",
		  basename(argv[0]),outfilename);
	  fflush(stderr);
	  exit(1);
	}
      }
      break;
    }
  }
  argidx = optind;
  while(args_processed < args_expected) {
    if (argidx >= argc) {
      /* missing arguments */
      printf("%s: missing argument(s)\n",basename(argv[0]));
      usage(argc,argv);
      exit(1);
    }
    switch(args_processed) {
    case 0:
      infilename = argv[argidx];
      if (strcmp(infilename,"-") == 0) {
	infd = 0;
      } else {
	infd = open(infilename,O_RDONLY);
	if (infd == -1) {
	  fprintf(stderr,"%s: unable to open file \"%s\" for input\n",
		  basename(argv[0]),infilename);
	  fflush(stderr);
	  exit(1);
	}
      }
      break;
    }
    argidx++;
    args_processed++;
  }
  if (argidx < argc) {
    /* extra arguments */
    printf("%s: extra argument(s)\n",basename(argv[0]));
    usage(argc,argv);
    exit(1);
  }
  return;
}

