%{
/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * snapparse.y : yacc parser for SNAP assembly language
 *
 * $Id$
 */

#include "config.h"
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <netinet/in.h>
#include "bytecode.h"
#include "dyncheck.h"
#include "packet.h"
#include "io.h"
#include "memalloc.h"
#include "labels.h"
#include "d_printf.h"
#include "list.h"

extern packet_t *p;

#define CHECK_CODE_OVERFLOW(p)				\
if ((p)->pc >= (p)->code_max) {			\
  fprintf(stderr,"%s:%d: code overflow\n",	\
	  __FILE__,__LINE__);			\
  fflush(stderr);				\
  exit(1);					\
}

int newho(buffer_t *b,char *file,int line);
int newtup(list_t *vlist, char *file, int line);

#ifdef SMALL_INSTRS
OPCODE_T refine_op(OPCODE_T op, TAG_T vtag);
#endif

void yyerror(char *s);

/* literal values from the lexer */
extern int value_int;
extern uint32 value_addr;
extern buffer_t value_str;
extern int value_exc;
extern float32 value_float;

extern int noop;

static char *label = NULL;

%}
%union {
  void *ptr;
  int integer;
}

%token T_INTV T_ADDRV T_STRV T_EXCV T_FLOATV
%token T_SRC T_DST T_RB T_SPORT T_DATA T_MAIN
%token T_EXIT T_PUSH T_POP T_POPI T_PULL T_EQ T_EQI T_PAJ T_TPAJ T_BEZ T_BNE
%token T_STORE T_JI T_NEQ T_NEQI
%token T_ADD T_ADDI T_SUB T_SUBI T_MULT T_MULTI T_DIV T_DIVI T_MOD T_MODI
%token T_NEG T_NOT T_LNOT T_AND T_ANDI T_OR T_ORI T_LSHL T_LSHLI T_RSHL 
%token T_RSHLI T_RSHA T_RSHAI T_SNET T_SNETI T_BCAST T_BCASTI T_ISX T_GETRB
%token T_GETSRC T_GETDST T_GETSPT T_HERE T_ISHERE T_ROUTE 
%token T_SEND T_HOP T_FORW T_FORWTO T_SENDCK T_HOPCK
%token T_DEMUX T_DEMUXI T_PRINT
%token T_PLUS T_MINUS T_LABEL T_LABELV T_PC
%token T_MKTUP T_LEN T_NTH T_ISTUP T_LPAREN T_RPAREN T_COMMA
%token T_SVCV T_CALLS
%token T_GT T_GEQ T_LT T_LEQ T_GTI T_GEQI T_LTI T_LEQI

%type <ptr> instr value value_list
%type <integer> instrs int_expr opint op1 op0 opaddr opstr
%left T_PLUS T_MINUS

%start instrs

%%
instrs: instr instrs    { $$ = $2; }
      | /* EOF */       { $$ = 0; }
      ;

instr:	op0 	        { CHECK_CODE_OVERFLOW(p);
			  SET_OP(*p->pc,$1);
			  p->pc++;
			  noop++;
			}
      | opint int_expr  { CHECK_CODE_OVERFLOW(p);
                          if (($2 > MAX_VINT) || ($2 < MIN_VINT)) {
			    fprintf(stderr,
			      "%s:%d: int literal %d out of bounds (%d,%d)\n",
			      __FILE__,__LINE__,$2,MIN_VINT,MAX_VINT);
			    fflush(stderr);
			    exit(1);
			  }
			  if (label != NULL) {
			    register_label_use(label,p->pc);
			    label = NULL;
			  }
			  SET_OP(*p->pc,$1);
			  SET_LIT(*p->pc,INTV,$2);
			  p->pc++;
                        }
      | opaddr T_ADDRV  { 
#ifdef ADDR_IN_HEAP
			  buffer_t b = 
			    { sizeof(value_addr), (char *)&value_addr };
			  int offs = newho(&b,__FILE__,__LINE__);
#endif
			  CHECK_CODE_OVERFLOW(p);
			  SET_OP(*p->pc,$1);
#ifdef ADDR_IN_HEAP
			  SET_LIT(*p->pc,ADDRV,offs);
#else
			  SET_LIT(*p->pc,ADDRV,value_addr);
#endif
			  p->pc++;
                        }
      | opstr T_STRV    { int offs = newho(&value_str,__FILE__,__LINE__);
			  free(value_str.s);
			  CHECK_CODE_OVERFLOW(p);
			  SET_OP(*p->pc,$1);
			  SET_LIT(*p->pc,STRV,offs);
			  p->pc++;
                        }
      | op1 value       { value_t *v = (value_t *)$2;
      			  int lit;
			  CHECK_CODE_OVERFLOW(p);
			  if (label != NULL) {
			    register_label_use(label,p->pc);
			    label = NULL;
			  }
#ifndef SMALL_INSTRS
			  SET_OP(*p->pc,$1);
#else
			  { OPCODE_T op = refine_op($1, GET_TAG(*v));
			    SET_OP(*p->pc,op); }
#endif
			  lit = GET_INT(*v); /* assumes the same as
						GET_* */
			  SET_LIT(*p->pc,GET_TAG(*v),lit);
			  p->pc++;
                        }
      | T_SRC T_ADDRV	{ p->hdr->src = value_addr; }
      | T_DST T_ADDRV	{ p->hdr->dst = value_addr; }
      | T_RB T_INTV	{ p->hdr->rb = htons((short)value_int); }
      | T_SPORT T_INTV  { p->hdr->sport = htons((short)value_int); }
      | T_MAIN		{ p->hdr->entry_point = htons(p->pc - p->code_min); }
      | T_LABEL		{ register_label_def(value_str.s,p->pc); }
      | T_DATA value    { value_t *v = (value_t *)$2;
			  if (p->sp >= p->stack_max) {
			    fprintf(stderr,"%s:%d: stack overflow\n",
			            __FILE__,__LINE__);
			    fflush(stderr);
			    exit(1);
			  }
			  if (label != NULL) {
			    register_label_use(label,p->sp);
			    label = NULL;
			  }
			  COPY_VAL(*p->sp,*v);
			  p->sp++;
			}
      ;

op0:	T_EXIT		{ $$ = (int)EXIT; }
      | T_POP           { $$ = (int)POP; }
      | T_EQ	        { $$ = (int)EQ; }
      | T_NEQ	        { $$ = (int)NEQ; }
      | T_ADD		{ $$ = (int)ADD; }
      | T_SUB		{ $$ = (int)SUB; }
      | T_MULT		{ $$ = (int)MULT; }
      | T_DIV		{ $$ = (int)DIV; }
      | T_MOD		{ $$ = (int)MOD; }
      | T_NEG		{ $$ = (int)NEG; }
      | T_NOT		{ $$ = (int)NOT; }
      | T_LNOT		{ $$ = (int)LNOT; }
      | T_AND		{ $$ = (int)AND; }
      | T_OR		{ $$ = (int)OR; }
      | T_LSHL		{ $$ = (int)LSHL; }
      | T_RSHL		{ $$ = (int)RSHL; }
      | T_RSHA		{ $$ = (int)RSHA; }
      | T_ISX		{ $$ = (int)ISX; }
      | T_GETRB		{ $$ = (int)GETRB; }
      | T_GETSRC	{ $$ = (int)GETSRC; }
      | T_GETDST	{ $$ = (int)GETDST; }
      | T_GETSPT	{ $$ = (int)GETSPT; }
      | T_HERE		{ $$ = (int)HERE; }
      | T_ISHERE	{ $$ = (int)ISHERE; }
      | T_ROUTE		{ $$ = (int)ROUTE; }
      | T_SEND		{ $$ = (int)SEND; }
      | T_SENDCK	{ $$ = (int)SENDCK; }
      | T_HOP		{ $$ = (int)HOP; }
      | T_HOPCK		{ $$ = (int)HOPCK; }
      | T_FORW		{ $$ = (int)FORW; }
      | T_FORWTO	{ $$ = (int)FORWTO; }
      | T_DEMUX		{ $$ = (int)DEMUX; }
      | T_PRINT		{ $$ = (int)PRINT; }
      | T_LEN		{ $$ = (int)LEN; }
      | T_ISTUP		{ $$ = (int)ISTUP; }
      | T_SNET		{ $$ = (int)SNET; }
      | T_BCAST		{ $$ = (int)BCAST; }
      | T_GT 		{ $$ = (int)GT; }
      | T_GEQ 		{ $$ = (int)GEQ; }
      | T_LT 		{ $$ = (int)LT; }
      | T_LEQ 		{ $$ = (int)LEQ; }
      ;

opint:	T_PULL 		{ $$ = (int)PULL; }
      | T_STORE         { $$ = (int)STORE; }
      | T_PAJ	        { $$ = (int)PAJ; }
      | T_TPAJ	        { $$ = (int)TPAJ; }
      | T_BEZ           { $$ = (int)BEZ; }
      | T_BNE		{ $$ = (int)BNE; }
      | T_POPI		{ $$ = (int)POPI; }
      | T_MODI		{ $$ = (int)MODI; }
      | T_ANDI		{ $$ = (int)ANDI; }
      | T_ORI		{ $$ = (int)ORI; }
      | T_LSHLI		{ $$ = (int)LSHLI; }
      | T_RSHLI		{ $$ = (int)RSHLI; }
      | T_RSHAI		{ $$ = (int)RSHAI; }
      | T_JI		{ $$ = (int)JI; }
      | T_MKTUP		{ $$ = (int)MKTUP; }
      | T_NTH		{ $$ = (int)NTH; }
      ;

opaddr:	T_BCASTI	{ $$ = (int)BCASTI; }
      | T_DEMUXI	{ $$ = (int)DEMUXI; }
      | T_SNETI		{ $$ = (int)SNETI; }
      ;

opstr:	T_SVCV		{ $$ = (int)SVCV; }
      |	T_CALLS 	{ $$ = (int)CALLS; }
      ;

op1:	T_PUSH          { $$ = (int)PUSH; }
      | T_EQI		{ $$ = (int)EQI; }
      | T_NEQI		{ $$ = (int)NEQI; }
      | T_ADDI		{ $$ = (int)ADDI; }
      | T_SUBI		{ $$ = (int)SUBI; }
      | T_MULTI		{ $$ = (int)MULTI; }
      | T_DIVI		{ $$ = (int)DIVI; }
      | T_GTI 		{ $$ = (int)GTI; }
      | T_GEQI 		{ $$ = (int)GEQI; }
      | T_LTI 		{ $$ = (int)LTI; }
      | T_LEQI 		{ $$ = (int)LEQI; }
      ;

value: int_expr		{ value_t *_v; memalloc(_v,value_t *,sizeof(value_t));
			  if (($1 > MAX_VINT) || ($1 < MIN_VINT)) {
			    fprintf(stderr,
			      "%s:%d: int literal %d out of bounds (%d,%d)\n",
			      __FILE__,__LINE__,$1,MIN_VINT,MAX_VINT);
			    fflush(stderr);
			    exit(1);
			  }
			  SET_TAG(*_v,INTV);
			  SET_INT(*_v,$1);
			  $$ = (void *)_v;
                        }
      | T_ADDRV	        { value_t *_v; 
#ifdef ADDR_IN_HEAP
			  buffer_t b = 
			    { sizeof(value_addr), (char *)&value_addr };
			  int offs = newho(&b,__FILE__,__LINE__);
#endif
      			  memalloc(_v,value_t *,sizeof(value_t));
                          SET_TAG(*_v,ADDRV);
#ifndef ADDR_IN_HEAP
			  SET_ADDR(*_v,value_addr,0 /* dummy arg */);
#else
			  SET_OFFS(*_v,offs);
#endif
			  $$ = (void *)_v;
			}
      | T_FLOATV	{ value_t *_v; 
#ifdef ADDR_IN_HEAP
			  buffer_t b = 
			    { sizeof(value_float), (char *)&value_float };
			  int offs = newho(&b,__FILE__,__LINE__);
#endif
      			  memalloc(_v,value_t *,sizeof(value_t));
                          SET_TAG(*_v,FLOATV);
#ifndef ADDR_IN_HEAP
			  SET_FLOAT(*_v,value_float,0 /* dummy arg */);
#else
			  SET_OFFS(*_v,offs);
#endif
			  $$ = (void *)_v;
			}
      | T_STRV          { value_t *_v; 
                          int offs = newho(&value_str,__FILE__,__LINE__);
			  free(value_str.s);
			  memalloc(_v,value_t *,sizeof(value_t));
			  SET_TAG(*_v,STRV);
			  SET_OFFS(*_v,offs);
			  $$ = (void *)_v;
                        }
      | T_EXCV		{ value_t *_v; 
			  if ((value_exc > MAX_VINT) || (value_exc < 0)) {
			    fprintf(stderr,
			      "%s:%d: int literal %d out of bounds (%d,%d)\n",
			      __FILE__,__LINE__,value_exc,0,MAX_VINT);
			    fflush(stderr);
			    exit(1);
			  }
			  memalloc(_v,value_t *,sizeof(value_t));
			  SET_TAG(*_v,EXCV);
			  SET_INT(*_v,value_exc);
			  $$ = (void *)_v;
			}
      | T_LPAREN value_list T_RPAREN
			{ value_t *_v; 
			  int offs = newtup((list_t *)$2,__FILE__,__LINE__);
			  free_list($2);
			  memalloc(_v,value_t *,sizeof(value_t));
			  SET_TAG(*_v,TUPLEV);
			  SET_OFFS(*_v,offs);
			  $$ = (void *)_v;
			}
      ;       

value_list : value T_COMMA value_list
		        { $$ = (void *)cons($1,(list_t *)$3); }
	   | value	{ $$ = (void *)cons($1,(list_t *)0); }
	   ;

int_expr: int_expr T_PLUS int_expr { $$ = $1 + $3; }
	| int_expr T_MINUS int_expr { $$ = $1 - $3; }
	| T_PC 		{ $$ = p->pc - p->code_min; }
	| T_INTV	{ $$ = value_int; }
        | T_LABELV	{ $$ = 0; label = value_str.s; }

%%
int newho(buffer_t *b,char *file,int line) {
  int sizeb = sizeof(heap_obj) + b->lenb;
  int offs;
  heap_obj *ho;
  
  if ((sizeb & 0x3) != 0) {
    sizeb = (sizeb & ~0x3) + 4;
  }
  if (((p->h_alloc_ptr + sizeb) > p->heap_max) ||
      (b->lenb > MAX_HEAPOBJ_SZ))  {
    fprintf(stderr,"%s:%d: heap alloc failed\n",file,line);
    fflush(stderr);
    exit(1);
  }
  ho = (heap_obj *)(p->h_alloc_ptr);
  offs = p->h_alloc_ptr - p->heap_min;

  d_printf(150,"%s:%d: alloc %d-string at @%#x, offs = %d\n",
	   __FILE__,__LINE__, b->lenb, p->h_alloc_ptr, offs);

  p->h_alloc_ptr = p->h_alloc_ptr + sizeb;
  bzero(ho,sizeb);
  ho->len = b->lenb;
  ho->flag = 0;
  memcpy((char *)(&(ho->s)),b->s,b->lenb);
  return(offs);
}

int newtup(list_t *list,char *file,int line) {
  int len = length_list(list);
  int lenb = len * sizeof(value_t);
  int sizeb = sizeof(heap_obj) + lenb;
  int offs;
  heap_obj *ho;
  value_t *varr;
  list_t *iter;
  
  assert ((sizeb & 0x3) == 0);
  if (((p->h_alloc_ptr + sizeb) > p->heap_max) ||
      (lenb > MAX_HEAPOBJ_SZ)) {
    fprintf(stderr,"%s:%d: heap alloc failed\n",file,line);
    fflush(stderr);
    exit(1);
  }
  ho = (heap_obj *)(p->h_alloc_ptr);
  offs = p->h_alloc_ptr - p->heap_min;

  d_printf(150,"%s:%d: alloc %d-tuple at @%#x, offs = %d\n",
	   __FILE__,__LINE__, lenb / sizeof(value_t),
	   p->h_alloc_ptr, offs);

  p->h_alloc_ptr = p->h_alloc_ptr + sizeb;
  bzero(ho,sizeb);
  ho->len = lenb;
  ho->flag = 1;

  varr = (value_t *)ho->s;
  for (iter = list; iter != NULL; iter = iter->next, varr++) {
    *varr = *((value_t *)iter->v);
  }
  return(offs);
}

#ifdef SMALL_INSTRS
OPCODE_T refine_op(OPCODE_T op, TAG_T vtag) {
  switch (op) {
  case PUSH:
    switch (vtag) {
    case INTV:
      op = PINT;
      break;
    case ADDRV: 
      op = PADDR;
      break;
    case STRV:
      op = PSTR;
      break;
    case EXCV:
      op = PEXC;
      break;
    case TUPLEV:
      op = PTUP;
      break;
    case FLOATV:
      op = PFLT;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d\n",__FILE__,__LINE__,vtag);
      exit(1);
    }
    break;
  case EQI:
    switch (vtag) {
    case INTV:
      op = EQINT;
      break;
    case ADDRV: 
      op = EQADR;
      break;
    case STRV:
      op = EQSTR;
      break;
    case EXCV:
      op = EQEXC;
      break;
    case TUPLEV:
      op = EQTUP;
      break;
    case FLOATV:
      op = EQFLT;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d\n",__FILE__,__LINE__,vtag);
      exit(1);
    }
    break;
  case NEQI:
    switch (vtag) {
    case INTV:
      op = NQINT;
      break;
    case ADDRV: 
      op = NQADR;
      break;
    case STRV:
      op = NQSTR;
      break;
    case EXCV:
      op = NQEXC;
      break;
    case TUPLEV:
      op = NQTUP;
      break;
    case FLOATV:
      op = NQFLT;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d\n",__FILE__,__LINE__,vtag);
      exit(1);
    }
    break;
  case ADDI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FADDI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for ADDI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  case SUBI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FSUBI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for SUBI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  case MULTI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FMULI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for MULTI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  case DIVI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FDIVI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for DIVI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  case GTI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FGTI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for GTI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  case GEQI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FGEQI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for GEQI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  case LTI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FLTI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for LTI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  case LEQI:
    switch (vtag) {
    case INTV:
      break;
    case FLOATV:
      op = FLEQI;
      break;
    default:
      fprintf(stderr,"%s:%d: invalid tag %d for LEQI\n",
	      __FILE__,__LINE__,vtag);
      break;
    }
    break;
  default:
    fprintf(stderr,"%s:%d: invalid opcode %d\n",__FILE__,__LINE__,op);
    exit(1);
  }
  return op;
}
#endif

void yyerror(char *s) {
  fprintf(stderr,s);
}
