/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * snapping.c : sample ping application for use with user-space SNAP
 *   daemons.
 *
 * $Id$
 */

#include <assert.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "d_printf.h"
#include "io.h"
#include "memalloc.h"
#include "packet.h"
#include "timers.h"

extern char *basename(const char *);

/* settable options */
int snapd_udp_portnum = 7777;
int my_udp_portnum  = 7800;
short starting_rb = 128;
int payload_sizeb = 64;
char *desthostname = NULL;
int num_packets = 1000;
int usec_delay = 1000000;
int verbose = 0;

/* sockets, addresses */
int sock = -1;
struct sockaddr_in destaddr = { 0, 0, { 0 } };
struct sockaddr_in localaddr = { 0, 0, { 0 } };
struct sockaddr_in myaddr = { 0, 0, { 0 } };

/* measurement */
int packet_sizeb = 0;
struct timeval *starts = NULL;
struct timeval *ends = NULL;
struct timeval *diffs = NULL;
int packets_sent = 0;

/* various buffers */
char sasfilename[128];
char sbcfilename[128];
char snapascmd[512];

void write_sas_file(void);
void parse_cmdline(int argc, char **argv);
void cleanup(void);
void sigint_handler(int ignore);

int main(int argc, char **argv) {
  /* sockets, addresses, buffers */
  char thishostname[128];
  struct hostent *desthostp = NULL;
  struct hostent *thishostp = NULL;
  int sbc_fd = -1;
  buffer_t pbuf = { 0, NULL };
  buffer_t inbuf = { 0, NULL };
  int *seqnum_spot_out = NULL;
  int *seqnum_spot_in = NULL;

  /* utility pointers for manipulating packet buffer */
  header_t *h = NULL;
  heap_obj *ho = NULL;
  
  /* set options */
  parse_cmdline(argc,argv);

  /* set up timing */
#ifndef TIMERS_OFF
  do_print_item_messages = 1;
  do_print_individual_timers = 1;
  print_flags = "10000000000001";
  print_flag_count = strlen(print_flags);
  init_all_timers();
  print_anti_timer(0,"total");
#endif /* !TIMERS_OFF */
  
  /* resolve addresses */
  if ((desthostp = gethostbyname(desthostname)) == NULL) {
    fprintf(stderr,"%s: unknown host %s\n",basename(argv[0]),
	    desthostname);
    exit(1);
  }
  assert(desthostp->h_addrtype == AF_INET);
  assert(desthostp->h_length == sizeof(unsigned int));
  memcpy(&destaddr.sin_addr.s_addr,desthostp->h_addr,desthostp->h_length);
  destaddr.sin_family = AF_INET;
  destaddr.sin_port = htons((short)snapd_udp_portnum);
  d_printf(10,"%s:%d: dest address 0x%08x\n",__FILE__,__LINE__,
	  destaddr.sin_addr.s_addr);
  
  if (gethostname(thishostname,sizeof(thishostname)) == -1) {
    perror("ping: gethostname");
    exit(1);
  }
  d_printf(10,"%s:%d: thishostname = %s\n",__FILE__,__LINE__,
	  thishostname);

  if ((thishostp = gethostbyname(thishostname)) == NULL) {
    perror("ping: gethostbyname thishost");
    exit(1);
  }
  assert(thishostp->h_addrtype == AF_INET);
  assert(thishostp->h_length == sizeof(unsigned int));
  if (localaddr.sin_addr.s_addr == 0) {
    memcpy(&localaddr.sin_addr.s_addr,thishostp->h_addr,thishostp->h_length);
  }
  memcpy(&myaddr.sin_addr.s_addr,thishostp->h_addr,thishostp->h_length);
  localaddr.sin_family = AF_INET;
  localaddr.sin_port = htons((short)snapd_udp_portnum);
  d_printf(10,"%s:%d: localaddr = 0x%08x\n",__FILE__,__LINE__,
	   localaddr.sin_addr.s_addr);

  myaddr.sin_family = AF_INET;
  myaddr.sin_port = htons((short)my_udp_portnum);
  myaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

  /* set up temporary file names */
  srand(time(NULL));
  sbcfilename[0] = '\0';
  sprintf(sasfilename,"/tmp/%s-%d",basename(argv[0]),rand());
  do {
    sprintf(sbcfilename,"/tmp/%s-%d",basename(argv[0]),rand());
  } while(!strcmp(sbcfilename,sasfilename));

  atexit(cleanup);
  signal(SIGINT,sigint_handler);

  /* now generate the assembly code for the ping packet */
  write_sas_file();

  /* call the assembler */
  sprintf(snapascmd,"snapas %s -o %s", sasfilename, sbcfilename);
  if (system(snapascmd) != 0) {
    fprintf(stderr,"%s: snapas command failed\n",basename(argv[0]));
    exit(1);
  }
  
  /* read in the binary packet */
  if ((sbc_fd = open(sbcfilename,O_RDONLY)) == -1) {
    perror("snapping: open sbcfile");
    exit(1);
  }
  packet_sizeb = file_to_str(sbc_fd,&pbuf);
  printf("%s:%d: binary ping packet is %d bytes\n",
	  __FILE__,__LINE__,packet_sizeb);
  fflush(stdout);

  /* prep the packet buffer */
  h = (header_t *)pbuf.s;
  h->src = localaddr.sin_addr.s_addr;
  h->dst = destaddr.sin_addr.s_addr;
  h->rb = htons(starting_rb);
  ho = (heap_obj *)((void *)(&(h->rest)) + /* skip header */
		    ntohs(h->code_sizeb));  /* skip code */

  assert(payload_sizeb >= 4);
  seqnum_spot_out = (int *)(&(ho->s)); /* blast the sequence number into
					  the first 4 bytes of the payload */
  
  /* set up receiving buffer too */
  memalloc(inbuf.s,char *,payload_sizeb);
  inbuf.lenb = payload_sizeb;
  seqnum_spot_in = (int *)inbuf.s;

  /* set up sockets and addressing */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("snapping: socket");
    exit(1);
  }
  if (bind(sock,&myaddr,sizeof(myaddr)) == -1) {
    perror("snapping: bind");
    exit(1);
  }

  /* set up timing data */
  memalloc(starts,struct timeval *,sizeof(struct timeval) * num_packets);
  memalloc(ends,struct timeval *,sizeof(struct timeval) * num_packets);
  memalloc(diffs,struct timeval *,sizeof(struct timeval) * num_packets);
  bzero(starts,sizeof(struct timeval) * num_packets);
  bzero(ends,sizeof(struct timeval) * num_packets);
  bzero(diffs,sizeof(struct timeval) * num_packets);

  /* main ping loop--send one, wait for response */
  for(packets_sent = 0; packets_sent < num_packets; packets_sent++) {
    int pnum_in;
    
    print_anti_timer(13,"ping");
    *seqnum_spot_out = packets_sent;
    if (gettimeofday(&(starts[packets_sent]),NULL) == -1) { 
      perror("snapping: gettimeofday(start)"); 
      fflush(stderr); 
    } 
    if (sendto(sock, pbuf.s, packet_sizeb, 0,
	       (struct sockaddr *)&localaddr,sizeof(localaddr)) !=
	packet_sizeb) {
      perror("snapping: sendto");
      fflush(stderr);
    }
    if (recvfrom(sock, inbuf.s, inbuf.lenb, 0, NULL, NULL) !=
	inbuf.lenb) {
      perror("snapping: recvfrom");
      fflush(stderr);
    }

    pnum_in = *seqnum_spot_in;
    d_printf(10,"%s:%d: seqnum_in = %d\n",__FILE__,__LINE__,pnum_in);

    if (pnum_in > packets_sent) {
      fprintf(stderr,"%s:%d: got a packet I didn't send!\n",
	      __FILE__,__LINE__);
      fflush(stderr);
      exit(1);
    }
    if (gettimeofday(&(ends[pnum_in]),NULL) == -1) {
      perror("snapping: gettimeofday(end)");
      fflush(stderr);
    }

    diffs[pnum_in].tv_sec = ends[pnum_in].tv_sec - starts[pnum_in].tv_sec;
    diffs[pnum_in].tv_usec = ends[pnum_in].tv_usec - starts[pnum_in].tv_usec;
    if (diffs[pnum_in].tv_usec < 0) {
      diffs[pnum_in].tv_sec--;
      diffs[pnum_in].tv_usec += 1000000;
    }
    if (verbose) {
      printf("ping %d: %ld.%06ld secs\n",packets_sent,
	     diffs[pnum_in].tv_sec,diffs[pnum_in].tv_usec);
      fflush(stdout);
    }
    print_timer(13,"ping");
    usleep(usec_delay);
  }
  return(0);
}

void write_sas_file(void) {
  FILE *f;
  char *progfmt = "\
forw     \
bne    4 \
push   1 \
getsrc   \
forwto   \
pop      \
demux    \
#data    |%d| \
#data    %d \
#data    0\n";

  if ((f = fopen(sasfilename,"w")) == NULL) {
    perror("write_sas_file: fopen");
    exit(1);
  }
  fprintf(f,progfmt,payload_sizeb,my_udp_portnum);
  fclose(f);
}

void print_final_stats(void) {
  int packets_lost = 0;
  int min_idx = -1;
  int max_idx = -1;
  double total_time = 0.0;
  int i;
  
  for(i=0; i<packets_sent; i++) {
    if (ends[i].tv_sec != 0 && ends[i].tv_usec != 0) {
      /* got packet back */
      if (min_idx < 0 ||
	  diffs[i].tv_sec < diffs[min_idx].tv_sec ||
	  (diffs[i].tv_sec == diffs[min_idx].tv_sec &&
	   diffs[i].tv_usec < diffs[min_idx].tv_usec)) {
	min_idx = i;
      }
      if (max_idx < 0 ||
	  diffs[i].tv_sec > diffs[max_idx].tv_sec ||
	  (diffs[i].tv_sec == diffs[max_idx].tv_sec &&
	   diffs[i].tv_usec > diffs[max_idx].tv_usec)) {
	max_idx = i;
      }
      total_time += (double)(diffs[i].tv_sec) +
	(double)(diffs[i].tv_usec) / 1000000.0;
    } else {
      packets_lost++;
    }
  }
  printf("PING: packet size = %d bytes, payload size = %d bytes\n",
	 packet_sizeb, payload_sizeb);
  printf("PING: %d packets sent, %d packets received, %.0f%% packet loss\n",
	 packets_sent, packets_sent - packets_lost,
	 (float)(packets_lost) / (float)(packets_sent));
  printf("PING: round-trip min/avg/max = %ld.%06ld/%.6f/%ld.%06ld s\n",
	 diffs[min_idx].tv_sec, diffs[min_idx].tv_usec,
	 total_time / (double)(packets_sent - packets_lost),
	 diffs[max_idx].tv_sec, diffs[max_idx].tv_usec);
  fflush(stdout);
}
    
void cleanup(void) {
  unlink(sasfilename);
  unlink(sbcfilename);
  close(sock);
  print_final_stats();
#ifndef TIMERS_OFF
  print_timer(0,"total");
  dump_all_timers();
#endif /* !TIMERS_OFF */
}

void sigint_handler(int ignore) {
  exit(0);
}

void usage(int argc, char **argv) {
  printf("usage: %s [-?v] [-p snapdport] [-a appport]\n",basename(argv[0]));
  printf("\t\t[-r resbound] [-b payload] [-d delay] [-l localaddr]\n");
  printf("\t\t[-n numpacks] dest\n");
  printf("\t-? : print this help\n");
  printf("\t-p : set the UDP port for snapd (default=7800)\n");
  printf("\t-a : set my UDP port (default=8000)\n");
  printf("\t-r : set the resource bound (default=128)\n");
  printf("\t-b : set the payload size in bytes (default=64)\n");
  printf("\t-d : set the microsecond delay between sends (default=1000000)\n");
  printf("\t-n : set number of packets to send (default=1000)\n");
  printf("\t-l : set local address (default=`hostname`)\n");
  printf("\t-v : verbose--print each ping time\n");
}

void parse_cmdline(int argc, char **argv) {
  char *opts = "?p:a:r:b:n:d:l:v";
  char c = '\0';
  int args_expected = 1;
  int args_processed = 0;
  int idx = 0;

  while((c = getopt(argc,argv,opts)) != EOF) {
    switch(c) {
    case '?':
      usage(argc,argv);
      exit(1);
    case 'p':
      snapd_udp_portnum = atoi(optarg);
      if (snapd_udp_portnum <= 0) {
	printf("%s: invalid port number %d\n",
	       basename(argv[0]), snapd_udp_portnum);
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'a':
      my_udp_portnum = atoi(optarg);
      if (my_udp_portnum <= 0) {
	printf("%s: invalid port number %d\n",
	       basename(argv[0]), my_udp_portnum);
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'r':
      starting_rb = atoi(optarg);
      if (starting_rb <= 0) {
	printf("%s: resource bound must be positive\n",basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'b':
      payload_sizeb = atoi(optarg);
      if (payload_sizeb < 0) {
	printf("%s: payload size must be nonnegative\n",basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      if (payload_sizeb < 4) {
	fprintf(stderr,"%s: raising payload size to 4 bytes\n",
		basename(argv[0]));
	fflush(stderr);
	payload_sizeb = 4;
      }
      break;
    case 'n':
      num_packets = atoi(optarg);
      if (num_packets <= 0) {
	printf("%s: must send a positive number of packets\n",
	       basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'l': {
      unsigned int b1,b2,b3,b4;
      unsigned int haddr;
      if (sscanf(optarg,"%d.%d.%d.%d",&b1,&b2,&b3,&b4) == 4) {
	haddr = (b1 << 24) | (b2 << 16) | (b3 << 8) | b4;
	localaddr.sin_addr.s_addr = htonl(haddr);
      }
      break;
    }
    case 'd':
      usec_delay = atoi(optarg);
      if (usec_delay <= 0) {
	printf("%s: must send a positive number of packets\n",
	       basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'v':
      verbose = 1;
      break;
    }
  }

  idx = optind;
  while(idx < argc) {
    switch(args_processed) {
    case 0:
      desthostname = argv[idx];
      break;
    default:
      printf("%s: too many argument(s)\n",basename(argv[0]));
      usage(argc,argv);
      exit(1);
    }
    args_processed++;
    idx++;
  }
  if (args_processed < args_expected) {
    printf("%s: missing argument(s)\n",basename(argv[0]));
    usage(argc,argv);
    exit(1);
  }
}

