/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * snaprecv.c : sample throughput receiver application for use with
 *   user-space SNAP daemons.
 *
 * $Id$
 */

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "d_printf.h"
#include "io.h"
#include "memalloc.h"
#include "packet.h"
#include "router.h"
#include "timers.h"

extern char *basename(const char *);

/* settable options */
int udp_port = 7777;

/* measurement */
int packets_recv = 0;
int packet_sizeb = 0;
int payload_sizeb = 0;
int first_packet = 1;
struct timeval start = { 0, 0 };
struct timeval end = { 0, 0 };

/* socket file descriptor, addresses */
int sock = -1;
struct sockaddr_in thisaddr = { 0, 0, { 0 } };

/* prototypes */
void parse_cmdline(int argc, char **argv);
void cleanup(void);
void sigint_handler(int ignore);
void print_final_stats(void);

int main(int argc, char **argv) {
  /* addresses, buffers, sockets */
  /* char thishostname[128];
     struct hostent *thishost = NULL; */
  buffer_t inbuf = { 0, NULL };

  /* packet buffer manipulation */
  header_t *h = NULL;
  heap_obj *ho = NULL;
  int *seqnum_spot = NULL;

  /* set up timing */
#ifndef TIMERS_OFF
  init_all_timers();
  print_anti_timer(0,"total");
#endif /* !TIMERS_OFF */

  /* read options */
  parse_cmdline(argc,argv);

  /* resolve addresses */
  thisaddr.sin_family = AF_INET;
  thisaddr.sin_port = htons((short)udp_port);
  thisaddr.sin_addr.s_addr = htonl(INADDR_ANY);

  atexit(cleanup);
  signal(SIGINT,sigint_handler);

  /* set up 10k packet buffer */
  memalloc(inbuf.s,char *,10 * 1024);
  inbuf.lenb = 10 * 1024;

  /* set up sockets and addressing */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("snaprecv: socket");
    exit(1);
  }
  if (bind(sock,(struct sockaddr *)&thisaddr,sizeof(thisaddr)) == -1) {
      perror("snaprecv: bind");
      exit(1);
  }

  /* ready message so we know to start the sender */
  printf("%s:%d: ready to receive\n",__FILE__,__LINE__);
  fflush(stdout);

  /* main receive loop */
  packets_recv = 0;
  first_packet = 1;
  while(1) {
    int bytes_got;
    if (!first_packet) {
      print_anti_timer(15,"snaprecv");
    }
    bytes_got = recvfrom(sock,inbuf.s,inbuf.lenb,0,NULL,NULL);
    if (bytes_got == -1) {
      perror("snaprecv: recvfrom");
      if (!first_packet) {
	print_timer(15,"snaprecv");
      }
      continue;
    }
    if (bytes_got == 1) {
      break;
    }
    if (start.tv_sec == 0 && start.tv_usec == 0) {
      if (gettimeofday(&start,NULL) == -1) {
	perror("snaprecv: gettimeofday(start)");
	exit(1);
      }
    }
    if (seqnum_spot == NULL) {
      packet_sizeb = bytes_got;
      h = (header_t *)inbuf.s;
      ho = (heap_obj *)((void *)(&(h->rest)) +
			ntohs(h->code_sizeb));
      payload_sizeb = ho->len;
      seqnum_spot = (int *)(&(ho->s));
    }
    packets_recv++;
    if (!first_packet) {
      print_timer(15,"snaprecv");
    }
    first_packet = 0;
  }

  return(0);
}

void print_final_stats(void) {
  long packet_bytes_recv = 0;
  long payload_bytes_recv = 0;
  long delta_sec = 0;
  long delta_usec = 0;

  if (gettimeofday(&end,NULL) == -1) {
    perror("snaprecv: gettimeofday(end)");
    exit(1);
  }
  packet_bytes_recv = packet_sizeb * packets_recv;
  payload_bytes_recv = payload_sizeb * packets_recv;
  delta_sec = end.tv_sec - start.tv_sec;
  delta_usec = end.tv_usec - start.tv_usec;
  if (delta_usec < 0) {
    delta_sec--;
    delta_usec += 1000000;
  }
  printf("RECV: %d packets, %ld packet bytes, %ld payload bytes\n",
	 packets_recv, packet_bytes_recv, payload_bytes_recv);
  printf("RECV: total receiving time: %ld.%06ld\n",delta_sec,delta_usec);
  printf("RECV: b/w total = %f Mb/s, b/w useful = %f Mb/s\n",
	 (double)packet_bytes_recv * 8 / (double)(1024 * 1024) /
	 ((double)delta_sec + (double)delta_usec / 1000000),
	 (double)payload_bytes_recv * 8 / (double)(1024 * 1024) /
	 ((double)delta_sec + (double)delta_usec / 1000000));
}

void cleanup(void) {
  close(sock);
  print_final_stats();
#ifndef TIMERS_OFF
  print_timer(0,"total");
  dump_all_timers();
#endif /* !TIMERS_OFF */
}

void sigint_handler(int ignore) {
  exit(0);
}

void usage(int argc, char **argv) {
  printf("usage: %s [-?i] [-p portnum] [-n numpacks] [-t tflags]\n",
	 basename(argv[0]));
  printf("\t-? : print this help\n");
  printf("\t-i : print individual timers\n");
  printf("\t-p : set UDP port (default=7777)\n");
  printf("\t-n : set the total number of packets to receive (default=1000)\n");
  printf("\t-t : which timers to print ({0|1}+)\n");
}

void parse_cmdline(int argc, char **argv) {
  char *opts = "?ip:t:";
  char c = '\0';
  int args_expected = 0;
  int args_processed = 0;
  int idx = 0;

  while((c = getopt(argc,argv,opts)) != EOF) {
    switch(c) {
    case '?':
      usage(argc,argv);
      exit(1);
    case 'i':
      do_print_individual_timers = 1;
      break;
    case 'p':
      udp_port = atoi(optarg);
      if (udp_port <= 0) {
	printf("%s: invalid port number %d\n",
	       basename(argv[0]), udp_port);
	usage(argc,argv);
	exit(1);
      }
      break;
    case 't':
      print_flags = optarg;
      print_flag_count = strlen(optarg);
      break;
    }
  }

  idx = optind;
  while(idx < argc) {
    switch(args_processed) {
    default:
      printf("%s: too many argument(s)\n",basename(argv[0]));
      usage(argc,argv);
      exit(1);
    }
    args_processed++;
    idx++;
  }
  if (args_processed < args_expected) {
    printf("%s: missing argument(s)\n",basename(argv[0]));
    usage(argc,argv);
    exit(1);
  }
}


