/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * snapsend.c : sample throughput sender application for use with
 *   user-space SNAP daemons.
 *
 * $Id$
 */

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "d_printf.h"
#include "io.h"
#include "memalloc.h"
#include "packet.h"
#include "router.h"
#include "timers.h"

extern char *basename(const char *);

/* settable options */
int udp_port = 7777;
short starting_rb = 128;
int payload_sizeb = 64;
char *desthostname = NULL;
int num_packets = 1000;
int usec_delay = 0;
int spin_delay = 0;

/* measurement */
int packets_sent = 0;
int packet_sizeb = 0;
struct timeval start = { 0, 0 };
struct timeval end = { 0, 0 };
struct timezone dummy = { 0, 0 };

/* socket file descriptor, addresses */
int sock = -1;
struct sockaddr_in destaddr = { 0, 0, { 0 } };

/* various string buffers */
char sasfilename[128];
char sbcfilename[128];
char snapascmd[512];

/* prototypes */
void write_sas_file(void);
void parse_cmdline(int argc, char **argv);
void cleanup(void);
void sigint_handler(int ignore);
void print_final_stats(void);

int main(int argc, char **argv) {
  /* addresses, buffers, sockets */
  struct hostent *desthost = NULL;
  struct sockaddr_in hopaddr = { 0, 0, { 0 } };
  int sbc_fd = -1;
  buffer_t pbuf = { 0, NULL };

  /* packet buffer manipulation */
  header_t *h = NULL;
  heap_obj *ho = NULL;
  int *seqnum_spot = NULL;

  /* set up timing */
#ifndef TIMERS_OFF
  init_all_timers();
  print_anti_timer(0,"total");
#endif /* !TIMERS_OFF */

  /* read options */
  parse_cmdline(argc,argv);

  /* resolve addresses */
  { 
    unsigned int b1,b2,b3,b4;
    if (sscanf(desthostname,"%d.%d.%d.%d",&b1,&b2,&b3,&b4) == 4) {
      unsigned int hdest = b1 << 24;
      hdest |= b2 << 16;
      hdest |= b3 << 8;
      hdest |= b4;
      destaddr.sin_addr.s_addr = htonl(hdest);
    } else {
      if ((desthost = gethostbyname(desthostname)) == NULL) {
	herror("snapsend: gethostbyname");
	exit(1);
      }
      assert(desthost->h_addrtype == AF_INET);
      assert(desthost->h_length == sizeof(unsigned int));
      memcpy(&destaddr.sin_addr.s_addr,desthost->h_addr,desthost->h_length);
    }
    destaddr.sin_family = AF_INET;
    destaddr.sin_port = htons((short)udp_port);
    d_printf(10,"%s:%d: dest address 0x%08x\n",__FILE__,__LINE__,
	     destaddr.sin_addr.s_addr);
  }

  read_routes();		/* initialize routing table */

  /* set up temporary file names */
  srand(time(NULL));
  sbcfilename[0] = '\0';
  sprintf(sasfilename,"/tmp/%s-%d",basename(argv[0]),rand());
  do {
    sprintf(sbcfilename,"/tmp/%s-%d",basename(argv[0]),rand());
  } while(!strcmp(sbcfilename,sasfilename));

  atexit(cleanup);
  signal(SIGINT,sigint_handler);

  /* now generate the assembly code for the ping packet */
  write_sas_file();

  /* call the assembler */
  sprintf(snapascmd,"snapas %s -o %s", sasfilename, sbcfilename);
  if (system(snapascmd) != 0) {
    fprintf(stderr,"%s: snapas command failed\n",basename(argv[0]));
    exit(1);
  }

  /* read in the binary packet */
  if ((sbc_fd = open(sbcfilename,O_RDONLY)) == -1) {
    perror("snapsend: open sbcfile");
    exit(1);
  }
  packet_sizeb = file_to_str(sbc_fd,&pbuf);
  d_printf(10,"%s:%d binary transport packet is %d bytes\n",
	   __FILE__,__LINE__,packet_sizeb);
  
  /* prep the packet buffer */
  h = (header_t *)pbuf.s;
  h->dst = destaddr.sin_addr.s_addr;
  h->rb = htons(starting_rb);
  ho = (heap_obj *)((void *)(&(h->rest)) + /* skip header */
		    ntohs(h->code_sizeb));  /* skip code */
  assert(payload_sizeb >= 4);
  seqnum_spot = (int *)(&(ho->s)); /* blast the sequence number into
				      the first 4 bytes of payload */
  
  /* set up sockets and addressing */
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("snapsend: socket");
    exit(1);
  }
  { unsigned int *hop = NULL;
    hopaddr.sin_family = AF_INET;
    hopaddr.sin_port = htons((short)udp_port);
    hop = nexthop(destaddr.sin_addr.s_addr);
    if (hop == NULL) {
      fprintf(stderr,"%s:%d: unable to determine next hop to 0x%08x\n",
	      __FILE__,__LINE__,destaddr.sin_addr.s_addr);
      exit(1);
    }
    if (*hop == 0) {
      hopaddr.sin_addr.s_addr = destaddr.sin_addr.s_addr;
    } else {
      hopaddr.sin_addr.s_addr = *hop;
    }
  }

  /* main sending loop */
  if (gettimeofday(&start,&dummy) == -1) {
    perror("snapsend: gettimeofday(start)");
    exit(1);
  }
  for(packets_sent = 0; packets_sent < num_packets; packets_sent++) {
    print_anti_timer(14,"snapsend");
    *seqnum_spot = packets_sent;

    if (sendto(sock, pbuf.s, packet_sizeb, 0,
	       (struct sockaddr *)&hopaddr,sizeof(hopaddr)) !=
	packet_sizeb) {
      perror("snapsend: sendto");
      fflush(stderr);
    }
    print_timer(14,"snapsend");
    if (usec_delay > 0) {
      usleep(usec_delay);
    } else if (spin_delay > 0) {
      int i;
      for(i=0; i<spin_delay; i++);
    }
  }

  return(0);
}

void print_final_stats(void) {
  long packet_bytes_sent = 0;
  long payload_bytes_sent = 0;
  long delta_sec = 0;
  long delta_usec = 0;

  if (gettimeofday(&end,&dummy) == -1) {
    perror("snapsend: gettimeofday(end)");
    exit(1);
  }
  packet_bytes_sent = packet_sizeb * packets_sent;
  payload_bytes_sent = payload_sizeb * packets_sent;
  delta_sec = end.tv_sec - start.tv_sec;
  delta_usec = end.tv_usec - start.tv_usec;
  if (delta_usec < 0) {
    delta_sec--;
    delta_usec += 1000000;
  }
  printf("SENT: %d packets, %ld packet bytes, %ld payload bytes\n",
	 packets_sent, packet_bytes_sent, payload_bytes_sent);
  printf("SENT: total sending time: %ld.%06ld\n",delta_sec,delta_usec);
  printf("SENT: b/w total = %f Mb/s, b/w useful = %f Mb/s\n",
	 (double)packet_bytes_sent * 8 / (double)(1024 * 1024) /
	 ((double)delta_sec + (double)delta_usec / 1000000),
	 (double)payload_bytes_sent * 8 / (double)(1024 * 1024) /
	 ((double)delta_sec + (double)delta_usec / 1000000));
}

void write_sas_file(void) {
  FILE *f;
  char *progfmt = "\
forw        \
demux       \
#data  |%d| \
#data  %d\n";

  if ((f = fopen(sasfilename,"w")) == NULL) {
    perror("write_sas_file: fopen");
    exit(1);
  }
  fprintf(f,progfmt,payload_sizeb,udp_port);
  fclose(f);
}

void kill_recv(void) {
  char c = 'a';

  if (sendto(sock,&c,1,0,(struct sockaddr *)&destaddr,
	     sizeof(destaddr)) != 1) {
    perror("snapsend: sento(kill_recv)");
  }
}
      

void cleanup(void) {
  unlink(sasfilename);
  unlink(sbcfilename);
  kill_recv();
  close(sock);
  print_final_stats();
#ifndef TIMERS_OFF
  print_timer(0,"total");
  dump_all_timers();
#endif /* !TIMERS_OFF */
}

void sigint_handler(int ignore) {
  exit(0);
}

void usage(int argc, char **argv) {
  printf("usage: %s [-?i] [-p portnum] [-r resbound] [-b payload]\n",
	 basename(argv[0]));
  printf("\t\t[-u usecdelay] [-d spindelay] [-n numpacks] [-t tflags] dest\n");
  printf("\t-? : print this help\n");
  printf("\t-i : print individual timers\n");
  printf("\t-p : set UDP port (default=7777)\n");
  printf("\t-r : set initial resource bound (default=128)\n");
  printf("\t-b : set the payload size in bytes (default=64)\n");
  printf("\t-u : set the microsecond delay between sends (default=0)\n");
  printf("\t-d : set the spin delay between sends (default=0)\n");
  printf("\t-n : set the total number of packets to send (default=1000)\n");
  printf("\t-t : which timers to print ({0|1}+)\n");
}

void parse_cmdline(int argc, char **argv) {
  char *opts = "?ip:r:b:u:d:n:t:";
  char c = '\0';
  int args_expected = 1;
  int args_processed = 0;
  int idx = 0;

  while((c = getopt(argc,argv,opts)) != EOF) {
    switch(c) {
    case '?':
      usage(argc,argv);
      exit(1);
    case 'i':
      do_print_individual_timers = 1;
      break;
    case 'p':
      udp_port = atoi(optarg);
      if (udp_port <= 0) {
	printf("%s: invalid port number %d\n",
	       basename(argv[0]), udp_port);
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'r':
      starting_rb = atoi(optarg);
      if (starting_rb <= 0) {
	printf("%s: resource bound must be positive\n",basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'b':
      payload_sizeb = atoi(optarg);
      if (payload_sizeb < 0) {
	printf("%s: payload size must be nonnegative\n",basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      if (payload_sizeb < 4) {
	fprintf(stderr,"%s: raising payload size to 4 bytes\n",
		basename(argv[0]));
	fflush(stderr);
	payload_sizeb = 4;
      }
      break;
    case 'n':
      num_packets = atoi(optarg);
      if (num_packets <= 0) {
	printf("%s: must send a positive number of packets\n",
	       basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'u':
      usec_delay = atoi(optarg);
      if (usec_delay < 0) {
	printf("%s: must have a nonnegative delay\n",
	       basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      break;
    case 'd':
      spin_delay = atoi(optarg);
      if (spin_delay < 0) {
	printf("%s: must have a nonnegative delay\n",
	       basename(argv[0]));
	usage(argc,argv);
	exit(1);
      }
      break;
    case 't':
      print_flags = optarg;
      print_flag_count = strlen(optarg);
      break;
    }
  }

  idx = optind;
  while(idx < argc) {
    switch(args_processed) {
    case 0:
      desthostname = argv[idx];
      break;
    default:
      printf("%s: too many argument(s)\n",basename(argv[0]));
      usage(argc,argv);
      exit(1);
    }
    args_processed++;
    idx++;
  }
  if (args_processed < args_expected) {
    printf("%s: missing argument(s)\n",basename(argv[0]));
    usage(argc,argv);
    exit(1);
  }
}


