/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore, Michael Hicks,
 *                                 and Scott Nettles.
 *
 * timers.c : low-overhead timers based on the Pentium cycle counter
 *   for flexible performance profiling.
 *
 * $Id$
 */

#include <stdio.h>
#include <sys/time.h>
#include <assert.h>

#ifndef TIMERS_OFF
#ifndef USE_CYCLE_COUNTER
#error "this timing information will not work without the cycle counter"
#endif
#endif

#include "timers.h"
#include "clock.h"

/****************************************************************
 * Prototypes
 ****************************************************************/

extern unsigned long long rdtsr();

/****************************************************************
 * Globals
 ****************************************************************/

struct item_timingstuff {
  int count;
  unsigned long long cum_cycles;
  unsigned long long last_counter_value;
#ifndef LOW_OVERHEAD
  unsigned int last_count_value;
#endif
  int depth;
};

struct timert_t {
  struct item_timingstuff 	timer;
  char				name[64];
};

#ifndef LOW_OVERHEAD
#define ENTRY(idx,name) \
    {{0,0,0,0,0}, name}
#else
#define ENTRY(idx,name) \
    {{0,0,0,0}, name}
#endif

struct timert_t timerlist[] =
{
ENTRY(0, "total"),
ENTRY(1, "interp_packet"),
ENTRY(2, "daemon_loop"),
ENTRY(3, "udp_send_packet"),
ENTRY(4, "udp_get_packet"),
ENTRY(5, "recvfrom"),
ENTRY(6, "sendto"),
ENTRY(7, "unmarshal_packet"),
ENTRY(8, "marshal_packet"),
ENTRY(9, "unmarshal_code"),
ENTRY(10, "unmarshal_stack"),
ENTRY(11, "unmarshal_heap"),
ENTRY(12, "memalloc"),
ENTRY(13, "ping"),
ENTRY(14, "snapsend"),
ENTRY(15, "snaprecv"),
ENTRY(16, "nexthop"),
ENTRY(17, "udp_ishere"),
};

#define NUM_TIMERS (sizeof(timerlist)/sizeof(struct timert_t))

int do_print_item_messages = 1;
int do_print_individual_timers = 0;
int do_print_antitimers = 0;
char *print_flags = "";
int print_flag_count = 0;

/****************************************************************
 * Functions
 ****************************************************************/

static void 
dump_time (char *label,
	   char *label2,
	   struct item_timingstuff *rp)
{
  if ( rp->count > 0 ) {
    printf("cum %s%s | %d total: %Lu cycles; average: %2.8f s\n",
	   label2, label, rp->count, rp->cum_cycles,
	   ((double)(rp->cum_cycles / (unsigned long long)rp->count)) / 
	   (MACHINE_CLOCKRATE_MHZ * 1000000.0));
  }
}    

static void 
dump_ind_time (char *label,
	       char *label2,
	       struct item_timingstuff *rp)
{
    printf("ind %s%s | %d diff: %2.8f c; tot: %Lu c; average: %2.8f s\n",
	   label2, label, rp->count, 
#ifndef LOW_OVERHEAD
	   ((double)(rp->last_counter_value / 
		     (unsigned long long)rp->last_count_value)) /
	   (MACHINE_CLOCKRATE_MHZ * 1000000.0), 
#else
	   ((double)rp->last_counter_value / 
	    (MACHINE_CLOCKRATE_MHZ * 1000000.0)), 
#endif
	   rp->cum_cycles,
	   (rp->count != 0) ?
	   ((double)(rp->cum_cycles / (unsigned long long)rp->count)) / 
	   (MACHINE_CLOCKRATE_MHZ * 1000000.0) :
	   0.0);
}    

static void 
dump_ind_anti_time (char *label,
		    char *label2,
		    struct item_timingstuff *rp)
{
    printf("ind %s%s | %d cv: %Lu c; tot: %Lu c; average: %2.8f s\n",
	   label2, label, rp->count, 
	   rp->last_counter_value,
	   rp->cum_cycles,
	   (rp->count != 0) ?
	   ((double)(rp->cum_cycles / (unsigned long long)rp->count)) / 
	   (MACHINE_CLOCKRATE_MHZ * 1000000.0) :
	   0.0);
}    

void internal_print_anti_time(int index, char *string)
{
#ifndef LOW_OVERHEAD
  if (do_print_item_messages) {
    struct item_timingstuff *entry = &(timerlist[index].timer);
    entry->depth++;
    if (entry->depth == 1) {
      entry->last_count_value = entry->count;
      entry->last_counter_value = rdtsr();
      if (do_print_individual_timers)
	dump_ind_anti_time(timerlist[index].name,"beg ",entry);
    }    
  }
#else
  timerlist[index].timer.last_counter_value = rdtsr();
#endif
}  

void internal_print_time(int index, char *string)
{ 
#ifndef LOW_OVERHEAD
  if (do_print_item_messages) {
    struct item_timingstuff *entry = &(timerlist[index].timer);
    entry->depth--;
    entry->count++;
    if (entry->depth == 0) {
      entry->last_count_value = entry->count - entry->last_count_value;
      /* XXX check for overflow ? */
      entry->last_counter_value = rdtsr() - entry->last_counter_value;
      entry->cum_cycles += entry->last_counter_value;
      if (do_print_individual_timers) {
	dump_ind_time(timerlist[index].name,"end ",entry);
      }
    }
  }
#else
  struct item_timingstuff *entry = &(timerlist[index].timer);
  entry->cum_cycles += (rdtsr() - entry->last_counter_value);
  entry->count++;
#endif
}

void reset_timer(int i)
{
  timerlist[i].timer.count = 0;
  timerlist[i].timer.cum_cycles = 0;
  timerlist[i].timer.last_counter_value = 0;
#ifndef LOW_OVERHEAD
  timerlist[i].timer.last_count_value = 0;
#endif
  timerlist[i].timer.depth = 0;
}

void init_all_timers(void)
{
  int i;
  for (i = 0; i < NUM_TIMERS ; i++) {
    reset_timer(i);
  }
}

void dump_all_timers(void)
{
  int i, m = (print_flag_count < NUM_TIMERS ? print_flag_count : NUM_TIMERS);
  int tmp = do_print_individual_timers;

  if (print_flag_count > 0)
    printf("DUMPING CUMULATIVE TIMERS\n");

  do_print_individual_timers = 0;
  
  for (i = 0; i < m; i++) {
    if (print_flags[i] == '1') {
      dump_time(timerlist[i].name, "norm ", &(timerlist[i].timer));
    }
  }
  do_print_individual_timers = tmp;  
}



