/* snap-1.0. Copyright (C) 2000 by Jonathan T. Moore and Michael Hicks.
 *
 * udpnet.c : implements the networking functionality specified in the
 *   interface for udpnet.h. The interface was done in this way to
 *   facilitate an in-kernel implementation of SNAP without requiring
 *   too many changes to the interpreter source code.
 *
 * $Id$
 */

#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include "bytecode.h"
#include "d_printf.h"
#include "io.h"
#include "memalloc.h"
#include "timers.h"
#include "udpnet.h"

int udp_portnum = DEFAULT_UDP_PORT;
int sd;				/* socket descriptor */

/* information about the local node */
struct hostent *thishost = NULL;
struct in_addr thisaddr = { 0 };

void init_udpnet(void) {
  
  char hostname[128];		/* current node's hostname */
  struct sockaddr_in interp_udp_addr;
#ifndef NDEBUG
  char addrstr[16];
#endif /* !NDEBUG */

  if (gethostname(hostname,sizeof(hostname)) < 0) {
    /* perror("init_udpnet: gethostname"); */
    /* exit(1); */
    strcpy(hostname,"localhost");
  }
  d_printf(10,"%s:%d: hostname = %s\n",__FILE__,__LINE__,hostname);
  thishost = gethostbyname(hostname);
  if (thishost == NULL) {
    /* herror("init_udpnet: gethostbyname"); */
    /* exit(1); */
    memalloc(thishost,struct hostent *,sizeof(struct hostent));
    thishost->h_addrtype = AF_INET;
    thishost->h_length = sizeof(unsigned int);
    thishost->h_name = hostname;
    memalloc(thishost->h_addr_list,char **,sizeof(char *));
    memalloc(thishost->h_addr_list[0],char *,sizeof(unsigned int));
    thishost->h_addr_list[0][0] = 127;
    thishost->h_addr_list[0][1] = 0;
    thishost->h_addr_list[0][2] = 0;
    thishost->h_addr_list[0][3] = 1;
  }
  assert(thishost->h_addrtype == AF_INET);
  assert(thishost->h_length == sizeof(unsigned int));
  if (thisaddr.s_addr == 0) {
    memcpy(&(thisaddr.s_addr),(thishost->h_addr_list[0]),sizeof(unsigned int));
  }
#ifndef NDEBUG
  sprintf_addr(addrstr,thisaddr.s_addr);
  d_printf(10,"%s:%d: local addr = %s\n",__FILE__,__LINE__,addrstr);
#endif /* !NDEBUG */
  
  if ((sd = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
    perror("init_udpnet: socket");
    exit(1);
  }
  
  interp_udp_addr.sin_family = AF_INET;
  interp_udp_addr.sin_port = udp_portnum;
  interp_udp_addr.sin_addr.s_addr = INADDR_ANY;
  
  if (bind(sd,
	   (struct sockaddr *)&interp_udp_addr,
	   sizeof(interp_udp_addr)) < 0) {
    perror("init_udpnet: bind");
    exit(1);
  }
  d_printf(10,"%s:%d: socket bound on port %d\n",
	  __FILE__,__LINE__,udp_portnum);
}

void close_udpnet(void) {
  d_printf(10,"%s:%d: closing socket\n",__FILE__,__LINE__);
  close(sd);
}

int udp_ishere(unsigned int addr) {
  int idx;
  char *curr_addr;
  int _result = 0;

  print_anti_timer(17,"udp_ishere");
  if (addr == thisaddr.s_addr) {
    print_timer(17,"udp_ishere");
    return(1);
  }
  assert(thishost != NULL);
  assert(thishost->h_length == sizeof(unsigned int));
  idx = 0;
  curr_addr = thishost->h_addr_list[0];
  while(_result == 0 && curr_addr != NULL) {
    if (addr == *(unsigned int *)curr_addr) {
      _result = 1;
      break;
    }
    idx++;
    curr_addr = thishost->h_addr_list[idx];
  }
  print_timer(17,"udp_ishere");
  return(_result);
}

void udp_send_packet(packet_t *p, int stack_amt, unsigned int dest) {
  buffer_t outbuf;
  struct sockaddr_in dest_addr;
#ifndef NDEBUG
  char addrstr[16];
#endif /* !NDEBUG */

  print_anti_timer(3,"udp_send_packet");
  /* marshal up the packet into a buffer */
  marshal_packet(p,stack_amt,&outbuf);

  dest_addr.sin_family = AF_INET;
  dest_addr.sin_port = udp_portnum;
  dest_addr.sin_addr.s_addr = dest;

  print_anti_timer(6,"sendto");
  if (sendto(sd, outbuf.s, outbuf.lenb, 0,
	     (struct sockaddr *)&dest_addr, sizeof(dest_addr)) == -1) {
    perror("udp_send_packet: sendto");
    if (errno != ECONNREFUSED) {
      exit(1);
    }
  }
  print_timer(6,"sendto");
#ifndef NDEBUG
  sprintf_addr(addrstr,dest);
  d_printf(25,"%s:%d: packet send: dst=%s:%d, size=%d bytes\n", 
	  __FILE__,__LINE__,addrstr,udp_portnum,outbuf.lenb);
#endif /* !NDEBUG */
  print_timer(3,"udp_send_packet");
}

/* recycle the 10k buffer */
char *inbuf = NULL;
int inbuf_sizeb = 10 * 1024;

packet_t *udp_get_packet(void) {
  int packet_lenb;		/* how many bytes in packet */
  packet_t *p;			/* packet we return */
  fd_set rfds;

#ifndef NDEBUG
  struct sockaddr from_addr;
  int from_len = sizeof(from_addr);
  char addrstr[16];
#endif /* !NDEBUG */
  
  print_anti_timer(4,"udp_get_packet");
  if (inbuf == NULL) {
    memalloc(inbuf,char *,inbuf_sizeb);
  }
  
  /* block until data ready to be read from socket -- allows us to
     time the recvfrom system call more directly */
  FD_ZERO(&rfds);
  FD_SET(sd, &rfds);
  if (select(sd+1, &rfds, NULL, NULL, NULL) == -1) {
    perror("udp_get_packet: select");
    exit(1);
  }
  print_anti_timer(5,"recvfrom");
#ifndef NDEBUG
  packet_lenb = recvfrom(sd,inbuf,inbuf_sizeb,0,&from_addr,&from_len);
#else
  packet_lenb = recvfrom(sd,inbuf,inbuf_sizeb,0,NULL,NULL);
#endif /* !NDEBUG */
  print_timer(5,"recvfrom");
  while (packet_lenb == -1 && errno == ECONNREFUSED) {
    print_anti_timer(5,"recvfrom");
#ifndef NDEBUG
    packet_lenb = recvfrom(sd,inbuf,inbuf_sizeb,0,&from_addr,&from_len);
#else
    packet_lenb = recvfrom(sd,inbuf,inbuf_sizeb,0,NULL,NULL);
#endif /* !NDEBUG */
    print_timer(5,"recvfrom");
  }
  if (packet_lenb == -1) {
    perror("udp_get_packet: recvfrom");
    exit(1);
  }
  assert(from_addr.sa_family == AF_INET);
#ifndef NDEBUG
  sprintf_addr(addrstr,((struct sockaddr_in *)&from_addr)->sin_addr.s_addr);
  d_printf(25,"%s:%d: packet recv: src=%s:%d, size=%d bytes\n",
	   __FILE__, __LINE__, addrstr, udp_portnum, packet_lenb); 
#endif /* !NDEBUG */
  p = unmarshal_packet(inbuf,packet_lenb,inbuf_sizeb);
  print_timer(4,"udp_get_packet");
  return(p);
}

